# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
>>> import locale, gettext
>>> gettext.NullTranslations().install()
>>> a = MusicalPitch.new_from_notename("g")
>>> b = MusicalPitch.new_from_notename("f")
>>> print(a - b)
2
>>> print((a - 2).get_octave_notename())
f
>>> print((a + 3).get_octave_notename())
ais
>>> print(a < b)
0
>>> print(a > b)
1
>>> print(MusicalPitch.new_from_int(55) == a)
1
>>> print(MusicalPitch.new_from_notename(a.get_notename()) == a)
1
>>> print(a.clone() == a, id(a) == id(a.clone()))
True False
>>> a=MusicalPitch()
>>> print(a.m_octave_i == a.m_accidental_i == a.m_octave_i == 0)
1
>>> print(a.semitone_pitch())
48
>>> print(a.get_octave_notename())
c
>>> print((a+2).get_octave_notename())
d
>>> print((2+a).get_octave_notename())
d
>>> print(MusicalPitch.new_from_notename("des'").get_notename())
des
>>> print(MusicalPitch.new_from_int(50).semitone_pitch())
50
>>> print(MusicalPitch.new_from_int(50).get_notename())
d
>>> n=MusicalPitch.new_from_notename("fis,")
>>> print(n.get_user_octave_notename())
f#,
>>> n=MusicalPitch.new_from_notename("b,,")
>>> print(n.get_octave_notename())
b,,
>>> print(n.get_user_octave_notename())
b,,
>>> gettext.translation('solfege', './share/locale/', languages=['nb_NO']).install()
>>> print(n.get_octave_notename())
b,,
>>> print(n.get_user_octave_notename())
<sub>1</sub>H
>>> print(n.get_user_notename())
h
>>> print(_("Close"))
Lukk
>>> n = MusicalPitch()
>>> n.set_from_notename("d'")
>>> print(n.get_octave_notename())
d'
"""

import logging
import random

from solfege.mpd import _exceptions

# The following are here so that the strings are caught by pygettext
_("notename|c")
_("notename|cb")
_("notename|cbb")
_("notename|c#")
_("notename|cx")
_("notename|d")
_("notename|db")
_("notename|dbb")
_("notename|d#")
_("notename|dx")
_("notename|e")
_("notename|eb")
_("notename|ebb")
_("notename|e#")
_("notename|ex")
_("notename|f")
_("notename|fb")
_("notename|fbb")
_("notename|f#")
_("notename|fx")
_("notename|g")
_("notename|gb")
_("notename|gbb")
_("notename|g#")
_("notename|gx")
_("notename|a")
_("notename|ab")
_("notename|abb")
_("notename|a#")
_("notename|ax")
_("notename|b")
_("notename|bb")
_("notename|bbb")
_("notename|b#")
_("notename|bx")


class InvalidNotenameException(_exceptions.MpdException):

    def __init__(self, n):
        _exceptions.MpdException.__init__(self)
        self.m_notename = n

    def __str__(self):
        return _("Invalid notename: %s") % self.m_notename


class MusicalPitch:
    LOWEST_STEPS = -28
    HIGHEST_STEPS = 47
    notenames = ('c', 'cis', 'd', 'dis', 'e', 'f', 'fis',
                 'g', 'gis', 'a', 'ais', 'b')
    natural_notenames = ('c', 'd', 'e', 'f', 'g', 'a', 'b')
    sharp_notenames = ('cis', 'dis', 'fis', 'gis', 'ais')

    def clone(self):
        r = MusicalPitch()
        r.m_octave_i = self.m_octave_i
        r.m_notename_i = self.m_notename_i
        r.m_accidental_i = self.m_accidental_i
        return r

    def new_from_notename(n):
        assert isinstance(n, str)
        r = MusicalPitch()
        r.set_from_notename(n)
        return r
    new_from_notename = staticmethod(new_from_notename)

    def new_from_int(i):
        assert isinstance(i, int)
        r = MusicalPitch()
        r.set_from_int(i)
        return r
    new_from_int = staticmethod(new_from_int)

    def __init__(self):
        """
         c,,,, is lowest: m_octave_i == -4, steps() == -28
         g'''''' is highest: m_octave_i = 6, steps() == 46
        """
        self.m_octave_i = self.m_accidental_i = self.m_notename_i = 0

    def transpose_by_musicalpitch(self, P):
        """Silly function used by mpd/parser.py and company
        (d') transposes up one major second.
        """
        tra = P.semitone_pitch() - 60
        old_p = self.semitone_pitch()
        self.m_notename_i = self.m_notename_i + P.m_notename_i
        self.m_accidental_i = self.m_accidental_i + P.m_accidental_i
        if self.m_notename_i > 6:
            self.m_notename_i = self.m_notename_i - 7
            self.m_octave_i = self.m_octave_i + 1
        self.m_octave_i = self.m_octave_i + P.m_octave_i - 1
        if self.semitone_pitch() - old_p < tra:
            self.m_accidental_i = self.m_accidental_i + 1
        elif self.semitone_pitch() - old_p > tra:
            self.m_accidental_i = self.m_accidental_i - 1
        self.sanitate_accidental()
        return self

    def sanitate_accidental(self):
        """
        Make use self.m_accidental_i is some of the values -2, -1, 0, 1, 2
        It can be out of this range if the musicalpitch has been transposed.
        This function will change notenames like gisisis, where m_accidental_i
        is 3 to ais where m_accidental_i is 1
        """
        if not -3 < self.m_accidental_i < 3:
            p = self.semitone_pitch()
            self.set_from_int(p)

    def enharmonic_flip(self):  # FIXME find proper name.
        """
        Change the notename, so that gis becomes aes.
        What about d, should it be cisis or eeses??

        his,  c deses
        cisis d deses
        disis e fes
        eis   f geses
        fisis g aeses
        gisis a beses
        aisis b ces'

        cis des
        dis es
        fis ges
        gis aes
        ais bes
        """
        if self.m_accidental_i == 1 and self.m_notename_i < 6:
            self.m_accidental_i = -1
            self.m_notename_i += 1

    def normalize_double_accidental(self):
        """
        Change the tone so that we avoid double accidentals.
        """
        if self.m_accidental_i == 2:
            if self.m_notename_i in (0, 1, 3, 4, 5):  # c d f g a
                self.m_notename_i += 1
                self.m_accidental_i = 0
            elif self.m_notename_i == 2:  # e
                self.m_notename_i = 3
                self.m_accidental_i = 1
            else:
                assert self.m_notename_i == 6  # b
                self.m_notename_i = 0
                self.m_accidental_i = 1
                self.m_octave_i += 1
        elif self.m_accidental_i == -2:
            if self.m_notename_i in (1, 2, 4, 5, 6):  # d e g a b
                self.m_notename_i -= 1
                self.m_accidental_i = 0
            elif self.m_notename_i == 3:  # f
                self.m_notename_i = 2
                self.m_accidental_i = -1
            else:
                assert self.m_notename_i == 0
                self.m_notename_i = 6
                self.m_accidental_i = -1
                self.m_octave_i -= 1

    def steps(self):
        return self.m_notename_i + self.m_octave_i * 7

    def semitone_pitch(self):
        return [0, 2, 4, 5, 7, 9, 11][self.m_notename_i] \
            + self.m_accidental_i + self.m_octave_i * 12 + 48

    def pitch_class(self):
        return ([0, 2, 4, 5, 7, 9, 11][self.m_notename_i] + self.m_accidental_i) % 12

    def set_from_int(self, midiint):
        self.m_octave_i = (midiint - 48) // 12
        self.m_notename_i = {0: 0, 1: 0, 2: 1, 3: 1, 4: 2, 5: 3, 6: 3, 7: 4,
                             8: 4, 9: 5, 10: 5, 11: 6}[midiint % 12]
        self.m_accidental_i = midiint \
            - (self.m_octave_i + 4) * 12 \
            - [0, 2, 4, 5, 7, 9, 11][self.m_notename_i]

    def set_from_notename(self, notename):
        if not notename:
            raise InvalidNotenameException(notename)
        tmp = notename
        self.m_accidental_i = self.m_octave_i = 0
        while notename[-1] in ["'", ","]:
            if notename[-1] == "'":
                self.m_octave_i = self.m_octave_i + 1
            elif notename[-1] == ",":
                self.m_octave_i = self.m_octave_i - 1
            notename = notename[:-1]
        if notename.startswith('es'):
            notename = 'ees' + notename[2:]
        if notename.startswith('as'):
            notename = 'aes' + notename[2:]
        while notename.endswith('es'):
            self.m_accidental_i = self.m_accidental_i - 1
            notename = notename[:-2]
        while notename.endswith('is'):
            self.m_accidental_i = self.m_accidental_i + 1
            notename = notename[:-2]
        try:
            self.m_notename_i = ['c', 'd', 'e', 'f', 'g', 'a', 'b'].index(notename)
        except ValueError:
            raise InvalidNotenameException(tmp)

    def randomize(self, lowest, highest):
        """
        lowest and highest can be an integer, string or a MusicalPitch instance
        """
        assert type(lowest) == type(highest)
        if isinstance(lowest, str):
            lowest = MusicalPitch.new_from_notename(lowest).semitone_pitch()
        if isinstance(highest, str):
            highest = MusicalPitch.new_from_notename(highest).semitone_pitch()
        self.set_from_int(random.randint(int(lowest), int(highest)))
        return self

    def __radd__(self, a):
        return self + a

    def __add__(self, i):
        """
        MusicalPitch + integer = MusicalPitch
        MusicalPitch + Interval = MusicalPitch
        """
        if isinstance(i, int):
            v = self.semitone_pitch()
            if not 0 <= v + i < 128:
                raise ValueError
            return MusicalPitch.new_from_int(v + i)
        elif i.__class__.__name__ == 'Interval':  # isinstance(i, interval.Interval):
            if not 0 <= self.semitone_pitch() + i.get_intvalue() < 128:
                raise ValueError
            r = self.clone()
            _p = r.semitone_pitch()
            r.m_notename_i = r.m_notename_i + i.m_interval * i.m_dir
            r.m_octave_i = r.m_octave_i + r.m_notename_i // 7 + i.m_octave * i.m_dir
            r.m_notename_i = r.m_notename_i % 7
            _diff = r.semitone_pitch() - _p
            r.m_accidental_i = r.m_accidental_i + (i.get_intvalue() - _diff)
            # to avoid notenames like ciscisciscis :
            if r.m_accidental_i > 2:
                #                     c  d  f  g  a
                if r.m_notename_i in (0, 1, 3, 4, 5):
                    r.m_accidental_i -= 2
                else:
                    assert r.m_notename_i in (2, 6), r.m_notename_i
                    r.m_accidental_i -= 1
                r.m_notename_i = r.m_notename_i + 1
                if r.m_notename_i == 7:
                    r.m_notename_i = 0
                    r.m_octave_i = r.m_octave_i + 1
            if r.m_accidental_i < -2:
                r.m_accidental_i = r.m_accidental_i + 2
                r.m_notename_i = r.m_notename_i - 1
                if r.m_notename_i == -1:
                    r.m_notename_i = 6
                    r.m_octave_i = r.m_octave_i - 1
            if not 0 <= int(self) <= 127:
                raise ValueError
            return r
        else:
            raise _exceptions.MpdException("Cannot add %s" % type(i))

    def __sub__(self, i):
        """
        MusicalPitch - MusicalPitch = integer
        MusicalPitch - integer = MusicalPitch
        """
        if isinstance(i, MusicalPitch):
            return self.semitone_pitch() - i.semitone_pitch()
        assert isinstance(i, int)
        v = self.semitone_pitch()
        assert 0 <= v - i < 128
        return MusicalPitch.new_from_int(v - i)

    def __int__(self):
        return self.semitone_pitch()

    def __lt__(self, B):
        return self.semitone_pitch() < B.semitone_pitch()

    def __le__(self, B):
        return self.semitone_pitch() <= B.semitone_pitch()

    def __eq__(self, B):
        return self.semitone_pitch() == B.semitone_pitch()

    def __str__(self):
        return "(MusicalPitch %s)" % self.get_octave_notename()

    def get_user_notename(self):
        # xgettext:no-python-format
        return self._format_notename(_i("notenameformat|%(notename)s"))

    def get_user_octave_notename(self):
        # xgettext:no-python-format
        return self._format_notename(_i("notenameformat|%(notename)s%(oct)s"))

    def get_notename(self):
        return self._format_notename("%(utnotename)s")

    def get_octave_notename(self):
        return self._format_notename("%(utnotename)s%(oct)s")

    def _format_notename(self, format_string):
        """
        utnotename : untranslated notename, solfege-internal format.
        notename  : as the value translated in the po file
        notename2 : lowercase, but capitalized if below the tone c (as
                    "c" is defined internally in solfege.
        suboct :  '' (nothing) for c or higher
                  <sub>1</sub> for c,
                  <sub>2</sub> for c,,
        suboct2:  '' (nothing) for c, and higher
                  <sub>1</sub> for c,,
                  <sub>2</sub> for c,,,
        supoct:   '' (nothing) for tones lower than c'
                  <sup>1</sup> for c'
                  <sup>2</sup> for c'' etc.
        """
        assert -3 < self.m_accidental_i < 3, self.m_accidental_i
        utnotename = ['c', 'd', 'e', 'f', 'g', 'a', 'b'][self.m_notename_i]\
            + ['eses', 'es', '', 'is', 'isis'][self.m_accidental_i + 2]
        notename = "notename|" \
            + ['c', 'd', 'e', 'f', 'g', 'a', 'b'][self.m_notename_i]\
            + ['bb', 'b', '', '#', 'x'][self.m_accidental_i + 2]
        notename = _i(notename)
        if self.m_octave_i < 0:
            notename2 = notename.capitalize()
        else:
            notename2 = notename
        if self.m_octave_i > 0:
            octave = "'" * self.m_octave_i
        elif self.m_octave_i < 0:
            octave = "," * (-self.m_octave_i)
        else:
            octave = ""
        if self.m_octave_i < 0:
            suboct = "<sub>%s</sub>" % (-self.m_octave_i)
        else:
            suboct = ""
        if self.m_octave_i < -1:
            suboct2 = "<sub>%s</sub>" % (-self.m_octave_i - 1)
        else:
            suboct2 = ""
        if self.m_octave_i > 0:
            supoct = "<sup>%s</sup>" % (self.m_octave_i)
        else:
            supoct = ""
        D = {'utnotename': utnotename,
             'notename': notename,
             'notename2': notename2,
             'suboct': suboct,
             'suboct2': suboct2,
             'supoct': supoct,
             'oct': octave}
        try:
            return format_string % D
        except KeyError:
            logging.error("musicalpitch: Bad translation of notenameformat string")
            return "%(notename)s%(oct)s" % D



#!/usr/bin/python3

import os
import sys
import re

r = re.compile("(?P<left>\# Copyright \(C\) )(?P<years>(\d\d\d\d,\s)*(\d\d\d\d))(?P<right>\s+Tom Cato Amundsen)")


def do_file(filename):
    for enc in ('utf-8', 'iso-8859-1'):
        try:
            f = open(filename, 'r', encoding=enc)
            s = f.read()
        except UnicodeDecodeError:
            s = None
            f.close()
        if s:
            break
    if not s:
        print("failed file:", filename)
    m = r.search(s)

    def func(r):
        years = r.group('years')
        if '2016' not in years:
            years = "%s, 2016" % years
        return r.group('left') + years + r.group('right')
    s2 = r.sub(func, s)
    if s2 != s:
        with open(filename, 'w', encoding=enc) as f:
            f.write(s2)


for root, dirs, files in os.walk("."):
    if root in ('./mylint-tmpdir', './build-branch', './manual-po-tmp'):
        continue
    if root.startswith("./.git"):
        continue
    for f in files:
        if ((f.endswith(".py") or f == 'Makefile')
            or root == './exercises/standard/lesson-files'):
                    do_file(os.path.join(root, f))



import os
import sys
from subprocess import Popen

prefix =  os.path.split(os.path.dirname(os.path.abspath(sys.argv[0])))[0]

if len(sys.argv) > 1 and sys.argv[1] == 'srcdir':
    python = os.path.join(prefix, "solfege", "win32", "python", "python.exe")
    cmd = [python, "solfege.py"] + sys.argv[2:]
else:
    sys.path.append("../share/solfege")
    if '--debug' in sys.argv:
        p = "python.exe"
    else:
        p = "pythonw.exe"
    python = os.path.join(prefix, "python", p)
    cmd = [python, "solfege"] + sys.argv[1:]

from solfege import winlang

lang = winlang.win32_get_langenviron()
if lang and (lang != 'system default'):
    os.environ['LANGUAGE'] = lang

Popen(cmd + sys.argv[1:])



# vim: set fileencoding=utf-8 :
# Solfege - free ear training software
# Copyright (C) 2007, 2008, 2011, 2016 Tom Cato Amundsen
# License is GPL, see file COPYING


import unittest
import traceback
import codecs

from solfege.testlib import I18nSetup, outdir, TmpFileBase
from solfege.dataparser import *
import solfege.parsetree as pt
import os


class TestLexer(unittest.TestCase):

    def test_get_line(self):
        l = Lexer("""#comment1
#comment2
#comment3

var = 3
""", None)
        self.assertEqual(l.get_line(0), "#comment1")
        self.assertEqual(l.get_line(1), "#comment2")
        self.assertEqual(l.get_line(2), "#comment3")
        self.assertEqual(l.get_line(3), "")
        self.assertEqual(l.get_line(4), "var = 3")

    def test_scan(self):
        p = Dataparser()
        p._lexer = Lexer("\"string\" name 1.2 2 (", p)
        self.assertEqual(p._lexer.scan(STRING), "string")
        self.assertEqual(p._lexer.scan(NAME), "name")
        self.assertEqual(p._lexer.scan(FLOAT), "1.2")
        self.assertEqual(p._lexer.scan(INTEGER), "2")
        p._lexer = Lexer("1 2 3", p)
        try:
            p._lexer.scan(STRING)
        except DataparserException as e:
            self.assertEqual("(line 1): 1 2 3\n"
                              "          ^",
                              e.m_nonwrapped_text)
        else:
            self.fail("DataparserException not raised")

    def test_unable_to_tokenize(self):
        p = Dataparser()
        try:
            p._lexer = Lexer("question { a = 3} |!", p)
        except UnableToTokenizeException as e:
            self.assertEqual("(line 1): question { a = 3} |!\n"
                              "                            ^",
                              e.m_nonwrapped_text)
        else:
            self.fail("UnableToTokenizeException not raised")
        try:
            p._lexer = Lexer("x = 4\n"
                             "question { a = 3} |!", p)
        except UnableToTokenizeException as e:
            self.assertEqual("(line 1): x = 4\n"
                              "(line 2): question { a = 3} |!\n"
                              "                            ^",
                              e.m_nonwrapped_text)
        else:
            self.fail("UnableToTokenizeException not raised")

    def _test_encodings_delcar_not_first(self):
        """
        FIXME: I disabled this test because people suddenly started
        to report that UnableToTokenizeException was not raised.
        """
        s = '#\n#\n#vim: set fileencoding=iso-8859-1 : \n' \
            'name = "øæå" '
        f = codecs.open(os.path.join(outdir, "file1"), 'w', 'iso-8859-1')
        f.write(s)
        f.close()
        f = open(os.path.join(outdir, "file1"), 'r')
        s = f.read()
        f.close()
        self.assertRaises(UnableToTokenizeException,
            Lexer, s, Dataparser())

    def _test_missing_encoding_definition_iso88591(self):
        """
        FIXME: I disabled this test because people suddenly started
        to report that UnableToTokenizeException was not raised.
        We write a simple datafile in iso-8859-1 encoding, but does not
        add the encoding line. The dataparser will assume files are utf-8
        by default, and will fail to tokenize.
        """
        s = 'name = "øæå" '
        f = codecs.open(os.path.join(outdir, "file1"), 'w', 'iso-8859-1')
        f.write(s)
        f.close()
        f = open(os.path.join(outdir, "file1"), 'r')
        s = f.read()
        f.close()
        self.assertRaises(UnableToTokenizeException,
            Lexer, s, Dataparser())

    def test_X(self):
        s = r"""
question {
   music = music("\staff \stemUp  {
   \clef violin \key d \minor \time 4/4
    c4
    }a")
}
"""
        p = Dataparser()
        p._lexer = Lexer(s, p)


class TestDataParser(TmpFileBase):
    parserclass = Dataparser

    def assertRaisedIn(self, methodname):
        t = traceback.extract_tb(sys.exc_info()[2])
        self.assertEqual(t[-1][2], methodname)

    def test_exception_statement_1(self):
        try:
            self.do_file("b")
        except DataparserSyntaxError as e:
            self.assertRaisedIn('statement')
            self.assertEqual("(line 1): b\n" +
                              "           ^",
                              e.m_nonwrapped_text)
            self.assertEqual(e.m_token, ('EOF', None, 1, 0))
        else:
            self.fail("DataparserSyntaxError not raised")

    def test_exception_statement_2(self):
        try:
            self.do_file("a)")
        except DataparserSyntaxError as e:
            self.assertRaisedIn('statement')
            self.assertEqual("(line 1): a)\n" +
                              "           ^",
                              e.m_nonwrapped_text)
            self.assertEqual(e.m_token, (')', ')', 1, 0))
        else:
            self.fail("DataparserSyntaxError not raised")

    def test_exception_statement_3(self):
        try:
            self.do_file("""#comment
  XYZ
""")
        except DataparserSyntaxError as e:
            self.assertRaisedIn('statement')
            self.assertEqual("(line 1): #comment\n" +
                              "(line 2):   XYZ\n" +
                              "(line 3): \n" +
                              "          ^",
                              e.m_nonwrapped_text)
        else:
            self.fail("DataparserSyntaxError not raised")

    def test_exception_statement_4(self):
        try:
            self.do_file("""#comment
  A)
""")

        except DataparserSyntaxError as e:
            self.assertRaisedIn('statement')
            self.assertEqual("(line 1): #comment\n" +
                              "(line 2):   A)\n" +
                              "             ^",
                              e.m_nonwrapped_text)
        else:
            self.fail("DataparserSyntaxError not raised")

    def test_exception_assignment(self):
        try:
            self.do_file("question = 3")
        except AssignmentToReservedWordException as e:
            self.assertRaisedIn('assignment')
            self.assertEqual("(line 1): question = 3\n" +
                              "          ^",
                              e.m_nonwrapped_text)
        else:
            self.fail("AssignmentToReservedWordException not raised")

    def test_istr_translations_in_file1(self):
        p = self.do_file("""
         foo = "foo-C"
         foo[nb] = "foo-nb"
         question {
           var = "var-C"
           var[nb] = "var-nb"
         }
        """)
        self.assertEqual(p.tree[0].right.evaluate({}, {}),
                          'foo-C')
        self.assertEqual(p.tree[1].right.evaluate({}, {}),
                          'foo-nb')
        self.assertEqual(p.tree[2][0].right.evaluate({}, {}),
                          'var-C')
        self.assertEqual(p.tree[2][1].right.evaluate({}, {}),
                          'var-nb')
        self.assertTrue(isinstance(p.tree[2][0].right.evaluate({}, {}), istr))

    def test_istr_translations_in_file_lang_before_C(self):
        """
        The Dataparser will accept setting the translated var before the
        C locale. But the lesson file interpreter will report an
        error for this.
        """
        p = self.do_file("""
           foo[no] = "foo-no"
           foo = "foo-C"
        """)
        self.assertEqual(p.tree[0].left, "foo[no]")
        self.assertEqual(p.tree[0].right.m_value, "foo-no")
        self.assertEqual(p.tree[0].right.m_value.cval, "foo-no")
        self.assertEqual(p.tree[1].right.m_value, "foo-C")

    def test_i18n_list_fails(self):
        try:
            self.do_file('foo[no] = "foo-no", "blabla" ')
        except CanOnlyTranslateStringsException as e:
            self.assertEqual(e.m_nonwrapped_text,
               '(line 1): foo[no] = "foo-no", "blabla" \n'
               + '                    ^')
        else:
            self.fail("CanOnlyTranslateStringsException not raised")

    def test_i18n_int_fails(self):
        try:
            self.do_file('foo[no] = 8')
        except CanOnlyTranslateStringsException as e:
            self.assertEqual(e.m_nonwrapped_text,
               '(line 1): foo[no] = 8\n'
               + '                    ^')
        else:
            self.fail("CanOnlyTranslateStringsException not raised")

    def test_import(self):
        p = self.do_file("\n".join([
            "import progression_elements",
            "t = progression_elements.I",
        ]))
        self.assertEqual(len(p.tree), 2)
        self.assertTrue(isinstance(p.tree[0], pt.Assignment))
        self.assertTrue(isinstance(p.tree[0].right, pt.Program))
        self.assertTrue(isinstance(p.tree[1], pt.Assignment))
        self.assertTrue(isinstance(p.tree[1].right, pt.Identifier))

    def test_import_as(self):
        p = self.do_file("\n".join([
            "import progression_elements as p",
            "t = p.I",
            "question {",
            "   prog = p.I",
            "}"]))
        self.assertEqual(len(p.tree), 3)
        self.assertTrue(isinstance(p.tree[0], pt.Assignment))
        self.assertTrue(isinstance(p.tree[0].right, pt.Program))
        self.assertTrue(isinstance(p.tree[1].right, pt.Identifier))
        self.assertTrue(isinstance(p.tree[2][0].right, pt.Identifier))

    def test_pt_2(self):
        p = self.do_file("""header {
        module = idbyname
        help = "idbyname-intonation"
        title = _("Is the interval flat, in tune or sharp? %s cent wrong") % 10
        lesson_heading = _("Just interval: %s") % _("Minor Second") + " (16:15)"
        filldir = vertic
}
        """)
        for d in p.tree[0]:
            if d.left == 'module':
                self.assertEqual(d.right, 'idbyname')

    def test_nested_block(self):
        """
        As we can see, the Dataparser class and the parsetree code
        can handle nested blocks.
        """
        p = self.do_file("question { a = 4 subbl { b = 5} }")
        self.assertTrue(isinstance(p.tree[0], pt.Block))
        self.assertTrue(isinstance(p.tree[0][0], pt.Assignment))
        self.assertTrue(isinstance(p.tree[0][1], pt.Block))
        self.assertTrue(isinstance(p.tree[0][1][0], pt.Assignment))
        self.assertEqual(p.tree[0][1][0].right.evaluate({}, {}), 5)

    def test_named_block(self):
        p = self.do_file('element I { label = "label-I" } '
                      + 'element II { label = "label-II" }')
        self.assertEqual(len(p.tree), 2)
        self.assertEqual(len(p.tree[0]), 1)
        self.assertEqual(len(p.tree[1]), 1)
        self.assertTrue(isinstance(p.tree[0], pt.NamedBlock))
        self.assertEqual(p.tree[0][0].right.evaluate({}, {}), 'label-I')
        self.assertTrue(isinstance(p.tree[1], pt.NamedBlock))
        self.assertEqual(p.tree[1][0].right.evaluate({}, {}), 'label-II')


class TestIstr(I18nSetup):

    def test_musicstr(self):
        s = istr(r"\staff{ c e g }")
        self.assertTrue(isinstance(s, str))

    def test_add_translations1(self):
        # i18n.langs: nb_NO, nb, C
        # name = "Yes"
        # name[no] = "Ja"
        s = "Yes"
        s = istr(s)
        self.assertEqual(str(s), 'Yes')
        s = s.add_translation('nb', 'Ja')
        self.assertEqual(s, 'Ja')
        s = s.add_translation('nb_NO', 'Ja!')
        self.assertEqual(s, 'Ja!')

    def test_add_translations2(self):
        # i18n.langs: nb_NO, nb, C
        # name = "Yes"
        # name[no] = "Ja"
        s = "Yes"
        s = istr(s)
        self.assertEqual(s, 'Yes')
        s = s.add_translation('nb_NO', 'Ja!')
        self.assertEqual(s, 'Ja!')
        s = s.add_translation('nb', 'Ja')
        self.assertEqual(s, 'Ja!', "Should still be 'Ja!' because no_NO is preferred before no")

    def test_override_gettext(self):
        s = dataparser_i18n_func("major")
        self.assertEqual(s, "dur")
        self.assertEqual(s.cval, "major")
        s = s.add_translation('nb', "Dur")
        self.assertEqual(s, "Dur")

    def test_type(self):
        s = istr("jo")
        self.assertIsInstance(s, istr)
        self.assertIsInstance(s, str)


class TestEncodingSniffer(TmpFileBase):
    parserclass = Dataparser

    def test_file_not_found(self):
        self.assertRaises(FileNotFoundError, read_encoding_marker_from_file, "asdfasdf")
        self.assertEqual(
            read_encoding_marker_from_file("exercises/standard/lesson-files/progression-atte"),
            "iso-8859-1")
        self.assertEqual(
            read_encoding_marker_from_file("exercises/standard/lesson-files/chord-min-major"),
            None)

    def test_from_string(self):
        self.assertEqual(read_encoding_marker_from_string(
            "# vim: set fileencoding=findme\n"
            + "# line\n"
            + "line\n"), "findme")
        self.assertEqual(read_encoding_marker_from_string(
            "# comment\n"
            + "# vim: set fileencoding=findme\n"
            + "# line\nline\n"), "findme")
        # Test that the file encoding is only detecte if placed in one
        # of the first two lines
        self.assertEqual(read_encoding_marker_from_string(
            "line 1\n"
            + "line 2\n"
            + "# vim: set fileencoding=findme\n"
            + "# line\nline\n"), None)

    def test_utf8_in_first_line(self):
        self.assertEqual(read_encoding_marker_from_string(
            "# ソルフェージュ\n"
            + "# vim: set fileencoding=findme\n"
            + "# line\nline\n"), "findme")

    def test_empty_string(self):
        self.assertIsNone(read_encoding_marker_from_string(""))

    def test_empty_lines(self):
        self.assertIsNone(read_encoding_marker_from_string("\n\n\n"))
        self.add_file("\n\n\n\n\n\n", "empty-lines")
        self.assertIsNone(read_encoding_marker_from_file(
            os.path.join(self.tmpdir, "empty-lines")))

    def test_empty_file(self):
        self.add_file("", "empty-file")
        self.assertIsNone(read_encoding_marker_from_file(
            os.path.join(self.tmpdir, "empty-file")))

suite = unittest.makeSuite(TestLexer)
suite.addTest(unittest.makeSuite(TestDataParser))
suite.addTest(unittest.makeSuite(TestIstr))
suite.addTest(unittest.makeSuite(TestEncodingSniffer))



# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2006, 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import Gtk

from solfege import cfg
from solfege import gu
from solfege import mpd
from solfege import soundcard

MAX_VOLUME = 127.0


class MidiInstrumentMenu(Gtk.Menu):

    def __init__(self, callback):
        Gtk.Menu.__init__(self)
        self.m_callback = callback
        for x in range(len(soundcard.instrument_names)):
            if x % 8 == 0:
                menuitem = Gtk.MenuItem(soundcard.instrument_sections[x // 8])
                submenu = Gtk.Menu()
                self.append(menuitem)
                menuitem.set_submenu(submenu)
                menuitem.show()
            item = Gtk.MenuItem(soundcard.instrument_names[x])
            item.connect('activate', self.on_activate, x)
            submenu.append(item)
            item.show()
        self.show()

    def on_activate(self, menuitem, instrument):
        self.m_callback(instrument)


class nInstrumentSelector(Gtk.VBox, cfg.ConfigUtils):

    def __init__(self, exname, name, sizegroup):
        Gtk.VBox.__init__(self)
        cfg.ConfigUtils.__dict__['__init__'](self, exname)
        self.m_name = name
        hbox = gu.bHBox(self)
        hbox.set_spacing(gu.PAD_SMALL)

        self.g_button = Gtk.Button(
              soundcard.instrument_names[self.get_int(self.m_name)])
        self.g_button.connect('clicked', self.on_btnclick)
        hbox.pack_start(self.g_button, True, True, 0)
        g = Gtk.VolumeButton()
        g.props.value = self.get_int('%s_volume' % name) // MAX_VOLUME

        def ff(volumebutton, value):
            self.set_int('%s_volume' % name, int(value * MAX_VOLUME))
        g.connect('value-changed', ff)
        hbox.pack_start(g, False, False, 0)

        self.g_menu = MidiInstrumentMenu(self.on_instrument_selected)
        self.m_instrument = self.get_int('preferred_instrument')

        hbox = Gtk.HBox()
        hbox.set_spacing(6)
        self.pack_start(hbox, True, True, 0)

    def on_btnclick(self, *argv):
        self.g_menu.popup(None, None, None, None, 1, 0)

    def on_instrument_selected(self, instrument=None):
        self.set_int(self.m_name, instrument)
        self.g_button.get_children()[0].set_text(soundcard.instrument_names[instrument])
        self.m_instrument = instrument
        self.play_selected_instrument()

    def play_selected_instrument(self, _o=None):
        t = mpd.Track()
        t.set_patch(self.m_instrument)
        t.set_volume(self.get_int('%s_volume' % self.m_name))
        t.note(4, 60)
        soundcard.synth.play_track(t)

    def show(self):
        self.show_all()


def FramedInstrumentSelector(title, exname, varname, sizegroup):
    box = Gtk.HBox()
    box.set_spacing(6)
    label = Gtk.Label(label=title)
    label.set_alignment(0.0, 0.5)
    box.pack_start(label, False, False, 0)
    n = nInstrumentSelector(exname, varname, None)
    sizegroup.add_widget(label)
    n.show()
    box.pack_start(n, True, True, 0)
    return box


class InstrumentConfigurator(Gtk.VBox, cfg.ConfigUtils):

    def __init__(self, exname, num, labeltext):
        Gtk.VBox.__init__(self)
        #cfg.ConfigUtils.__init__(self, exname)
        cfg.ConfigUtils.__dict__['__init__'](self, exname)
        assert num in (2, 3)
        self.m_num = num
        self.g_override_default_instrument_checkbutton \
            = gu.nCheckButton(exname, 'override_default_instrument',
                labeltext,
                 callback=self.update_instrument_override)
        self.pack_start(self.g_override_default_instrument_checkbutton,
                        False, False, 0)
        hbox = gu.bVBox(self)
        hbox.set_spacing(gu.PAD_SMALL)

        sizegroup = Gtk.SizeGroup(Gtk.SizeGroupMode.HORIZONTAL)
        self.g_instrsel_high = FramedInstrumentSelector(_("Highest:"), exname, 'highest_instrument', sizegroup)
        hbox.pack_start(self.g_instrsel_high, False, False, 0)
        if num == 3:
            self.g_instrsel_middle = FramedInstrumentSelector(_("Middle:"),
                                            exname, 'middle_instrument', sizegroup)
            hbox.pack_start(self.g_instrsel_middle, False, False, 0)
        else:
            self.g_instrsel_middle = None
        self.g_instrsel_low = FramedInstrumentSelector(_("Lowest:"),
                                                exname, 'lowest_instrument', sizegroup)
        hbox.pack_start(self.g_instrsel_low, False, False, 0)
        self.update_instrument_override()

    def update(self):
        self.update_instrument_override()
        self.g_instrsel_high.update()
        if self.m_num == 3:
            self.g_instrsel_middle.update()
        self.g_instrsel_low.update()

    def update_instrument_override(self, _o=None):
        self.g_override_default_instrument_checkbutton.set_active(
                self.get_bool('override_default_instrument'))
        self.g_instrsel_high.set_sensitive(
               self.g_override_default_instrument_checkbutton.get_active())
        self.g_instrsel_low.set_sensitive(
               self.g_override_default_instrument_checkbutton.get_active())
        if self.g_instrsel_middle:
            self.g_instrsel_middle.set_sensitive(
               self.g_override_default_instrument_checkbutton.get_active())

    def show(self):
        self.show_all()



#!/usr/bin/python3
# Solfege - free ear training software
# Copyright (C) 2007, 2008, 2011, 2016 Tom Cato Amundsen
# License is GPL, see file COPYING

from __future__ import division

import codecs
import sys
import os

# Need this to run under debuild, but not from the shell. Strange...
sys.path.insert(0, ".")
import solfege.i18n
solfege.i18n.setup(".")

import solfege.frontpage as fp
from solfege.mpd.musicalpitch import MusicalPitch
from solfege.mpd.interval import Interval
from pathlib import Path

saveto = Path("exercises/standard/lesson-files")

header = """
header {
  lesson_id = "generated-hear-tones-%s"
  title = %s
  module = singanswer
  version = "3.9.2"
}
"""


ratio_dict = {
        'm2': (16 / 15, "16:15"),
        'M2': (9 / 8, "9:8"),
        'm3': (6 / 5, "6:5"),
        'M3': (5 / 4, "5:4"),
        'p4': (4 / 3, "4:3"),
        'p5': (3 / 2, "3:2"),
        'm6': (8 / 5, "8:5"),
        'M6': (5 / 3, "5:3"),
        'm7': (16 / 9, "16:9"),
        'M7': (15 / 8, "15:8"),
        'p8': (2, "2:1"),
        'm9': (1 + 16 / 15, "1+16:15"),
        'M9': (1 + 9 / 8, "1+9:8"),
        'm10': (1 + 6 / 5, "1+6:5"),
        'M10': (1 + 5 / 4, "1+5:4"),
        }

interval_names = list(ratio_dict.keys())
interval_names.sort(key=lambda i: Interval(i).get_intvalue())

# different number of cents the intonation questions should be wrong.
cents = (40, 30, 20, 15, 10, 8, 6, 5)


def write_file_copyright(f):
    f.write("# Solfege - free ear training software\n")
    f.write("# Copyright (C) 2007, 2008, 2011, 2016 Tom Cato Amundsen\n")
    f.write("# License is GPL, see file COPYING\n")
    f.write("# Generated by tools/generate_lessonfiles.py\n")


def question(f, b, c):
    low = 60
    tone_a = MusicalPitch.new_from_int(low)
    tone_b = MusicalPitch.new_from_int(low + b)
    tone_c = MusicalPitch.new_from_int(low + c)
    f.write('question {\n')
    f.write('  question_text = _("Sing the three tones")\n')
    f.write('  music = voice("< %s2 %s %s >")\n' % (
        (tone_a.get_octave_notename(),
         tone_b.get_octave_notename(),
         tone_c.get_octave_notename())))
    f.write('  answer = voice("%s8 %s %s")\n' % (
        (tone_a.get_octave_notename(),
         tone_b.get_octave_notename(),
         tone_c.get_octave_notename())))
    f.write('}\n')


def three_tones_less_than_octave(filename):
    """3 toner innenfor en oktav
    0 1 2-12
    0 2 3-12
    0 3 4-12
    0 10 11-12
    0 11 12
    a  b  c
    """
    f = filename.open("w")
    write_file_copyright(f)
    f.write("""header {
  lesson_id = "80cc940b-c47b-4264-a7b5-6cbd7a997bb0"
  title = _("Triads")
  module = singanswer
  version = "3.9.2"
}
""")
    for b in range(1, 12):
        for c in range(b + 1, 13):
            question(f, b, c)
    f.close()


def three_tones_larger_than_octave(filename):
    """3 toner innenfor en oktav
    0 1  13-24      13==cis  24=c
    0 2  13-24
    0 11 13-24
    0 12 13-24    c c cis-c

   #0 12 13-24
    0 13 14-24
    0 14 15-24
    0 22 23-24
    0 23    24
    a  b  c
    """
    f = filename.open("w")
    write_file_copyright(f)
    f.write("""header {
  lesson_id = "3f1c7305-9978-4fe3-9bb3-8c01ff776a08"
  title = _("Triads, difficult")
  module = singanswer
  version = "3.9.2"
}
""")
    for b in range(1, 13):
        for c in range(13, 25):
            question(f, b, c)
    for b in range(13, 24):
        for c in range(b + 1, 25):
            question(f, b, c)
    f.close()


def csound_intonation(mod, lesson_id):
    filename = saveto / ("csound-fifth-%s" % mod)
    f = filename.open("w")
    f.write("""# Solfege - free ear training software
# Copyright (C) 2004, 2005, 2011, 2016 Tom Cato Amundsen
# License is GPL, see file COPYING

header {{
    lesson_id = "{0}"
        module = idbyname
        help = "idbyname-intonation"
        title = _("Is the fifth flat, in tune or sharp? %s") % "({1})"
        filldir = vertic
        fillnum = 4
}}
""".format(lesson_id, mod))
    for s, d in enumerate((1, 9 / 8, 4 / 3, 3 / 2, 5 / 3, 15 / 8, 2)):
        for label, mult in (("intonation|flat", float(mod)), ("intonation|in tune", 1), ("intonation|sharp", 1 / float(mod))):
            f.write("""
question {{
 name = _i("{0}")
 set={1}
 csound(load("share/sinus.orc"), \"\"\"
    f1 0 4096 10 1 1.25 0.95 0.8 0.6 0.4 0.2 
    i1 0 1 {2:f}
    i1 + 1 {3:f}
 \"\"\")
}}
""".format(label, s, 220.0 * d, 220.0 * d * 3 / 2 * mult))
    f.close()


def iname_to_fn(s):
    return s.replace("m", "min").replace("M", "maj")


def csound_intonation2(harmonic, interval_name, cent):
    i = Interval(interval_name)
    if harmonic:
        hs = "harmonic-"
    else:
        hs = ""
    filename = saveto / ("csound-intonation-%s%s-%icent" % (hs, iname_to_fn(interval_name), cent))
    f = filename.open("w")
    f.write("""# Solfege - free ear training software
# Copyright (C) 2004, 2005, 2011, 2016 Tom Cato Amundsen
# License is GPL, see file COPYING

header {{
    lesson_id = "csound-intonation-{0}{1}-{2}cent"
        module = idbyname
        help = "idbyname-intonation"
        title = _("Is the interval flat, in tune or sharp? %s cent wrong") % {3}
        lesson_heading = _("Just interval: %s") % _("{4}") + " ({5})"
        filldir = vertic
}}
""".format(hs, iname_to_fn(interval_name), cent, cent,  # 0-3
           i.get_cname(), ratio_dict[i.get_cname_short()][1]))  # 4-5
    for c in range(12):
        hz = 220 * 2 ** (c / 12)
        hz2 = hz * ratio_dict[i.get_cname_short()][0]
        for label, mult in (("intonation|flat", -cent),
                            ("intonation|in tune", 0),
                            ("intonation|sharp", cent)):
            f.write("""
question {{
 name = _i("{0}")
 set={1}
 csound(load("share/sinus.orc"), \"\"\"
   f1 0 4096 10 1 1.25 0.95 0.8 0.6 0.4 0.2 
""".format(label, c))
            if harmonic:
                f.write('   i1 0 1 %f\n' % hz)
                f.write('   i1 0 1 %f\n' % (hz2 * 2**(mult / 1200)))
            else:
                f.write('   i1 0 1 %f\n' % hz)
                f.write('   i1 + 1 %f\n' % (hz2 * 2**(mult / 1200)))
            f.write(' """)\n')
            f.write("}\n")
    f.close()


def generate_csound_intonation():
    # We don't specify a lessonfile manager since we don't need
    # any of the methods that depend on it.
    page = fp.Page(u"Experimental CSound exercises", fp.Column())
    fileheader = fp.FileHeader(1, page)
    section = fp.LinkList(u"Melodic intervals")
    page[-1].append(section)
    for idx, cname in enumerate(interval_names):
        subpage = fp.Page(Interval(cname).get_cname().title(), fp.Column())
        section.append(subpage)
        sect = fp.LinkList(Interval(cname).get_cname().title())
        subpage[-1].append(sect)
        for cent in cents:
            csound_intonation2(False, cname, cent)
            sect.append("solfege:lesson-files/csound-intonation-%s-%icent" % (
                iname_to_fn(cname), cent))
    section = fp.LinkList(u"Harmonic intervals")
    col = fp.Column()
    page.append(col)
    page[-1].append(section)
    for idx, cname in enumerate(interval_names):
        subpage = fp.Page(Interval(cname).get_cname().title(), fp.Column())
        section.append(subpage)
        sect = fp.LinkList(Interval(cname).get_cname().title())
        subpage[-1].append(sect)
        for cent in cents:
            csound_intonation2(True, cname, cent)
            sect.append("solfege:lesson-files/csound-intonation-harmonic-%s-%icent" % (
                iname_to_fn(cname), cent))
    filename = saveto.parent / "csound-tree.txt"
    f = filename.open("w")
    fileheader.dump(f)
    f.close()

generate_csound_intonation()

for pct, lesson_id in (
        ("0.97", "b465c807-d7bf-4e3a-a6da-54c78d5b59a1"),
        ("0.98", "aa5c3b18-664b-4e3d-b42d-2f06582f4135"),
        ("0.99", "5098fb96-c362-45b9-bbb3-703db149a079"),
        ("0.995", "3b1f57e8-2983-4a74-96da-468aa5414e5e"),
        ("0.996", "a06b5531-7422-4ea3-8711-ec57e2a4ce22"),
        ("0.997", "e67c5bd2-a275-4d9a-96a8-52e43a1e8987"),
        ("0.998", "1cadef8c-859e-4482-a6c4-31bd715b4787")):
    csound_intonation(pct, lesson_id)
three_tones_less_than_octave(saveto / "hear-tones-triads")
three_tones_larger_than_octave(saveto / "hear-tones-triads-difficult")



# Solfege - free ear training software
# Copyright (C) 2007, 2008, 2011, 2016 Tom Cato Amundsen
# License is GPL, see file COPYING

import unittest
from solfege import mpd
from solfege.testlib import TmpFileBase
from solfege.mpd.track import Track, MidiEventStream
from solfege.mpd.rat import Rat
from solfege import lessonfile
from solfege import cfg


class TmpFile(TmpFileBase):
    parserclass = lessonfile.QuestionsLessonfile


class TestTrack(unittest.TestCase):

    def test_simple1(self):
        t = Track()
        t.note(4, 90, 127)
        self.assertEqual(
            list(MidiEventStream(t)),
            [('program-change', 0, 0),
             ('volume', 0, 100),
             ('note-on', 0, 90, 127),
             ('notelen-time', Rat(1, 4)),
             ('note-off', 0, 90, 127)])

    def test_1voice_setpatch(self):
        t = Track()
        t.note(4, 90, 127)
        t.set_patch(3)
        t.note(4, 91, 127)
        self.assertEqual(
            list(MidiEventStream(t)),
            [('program-change', 0, 0),
             ('program-change', 1, 3),
             ('volume', 0, 100),
             ('note-on', 0, 90, 127),
             ('notelen-time', Rat(1, 4)),
             ('note-off', 0, 90, 127),
             ('volume', 1, 100),
             ('note-on', 1, 91, 127),
             ('notelen-time', Rat(1, 4)),
             ('note-off', 1, 91, 127),
             ])

    def test_midi_overlap(self):
        t1 = Track()
        t1.note(8, 92, 121)
        t1.note(8, 90, 122)
        t1.note(8, 92, 123)
        t1.note(8, 94, 124)
        t2 = Track()
        t2.note(2, 90, 120)
        m = MidiEventStream(t1, t2)
        self.assertEqual("p0:0 p1:0 v0:100 n0:92 n0:90 d1/8 o92 v1:100 n1:90 d1/8 o90 n0:92 d1/8 o92 n0:94 d1/8 o94 o90", m.str_repr(1))

    def test_add(self):
        t1 = Track()
        t1.note(8, 92, 121)
        self.assertEqual("n92 d1/8 o92", t1.str_repr())
        t2 = Track()
        t2.note(2, 90, 120)
        self.assertEqual("n90 d1/2 o90", t2.str_repr())
        self.assertEqual("n92 d1/8 o92", t1.str_repr())
        self.assertEqual("n90 d1/2 o90", t2.str_repr())
        m1 = MidiEventStream(t1)
        m2 = MidiEventStream(t2)
        t3 = t1 + t2
        self.assertEqual("n92 d1/8 o92 n90 d1/2 o90", t3.str_repr())
        m3 = MidiEventStream(t3)
        self.assertEqual("n92 d1/8 o92", t1.str_repr())
        self.assertEqual("n90 d1/2 o90", t2.str_repr())
        # FIXME volume bug?
        self.assertEqual("p0:0 v0:100 n0:92 d1/8 o92", m1.str_repr(1))
        self.assertEqual("p0:0 v0:100 n0:90 d1/2 o90", m2.str_repr(1))
        self.assertEqual("p0:0 v0:100 n0:92 d1/8 o92 n0:90 d1/2 o90", m3.str_repr(1))


class TestMidiEventStream(TmpFile):

    def setUp(self):
        TmpFile.setUp(self)
        cfg.set_bool('config/override_default_instrument', True)

    def test_track1(self):
        t = Track()
        t.prepend_patch(33)
        t.note(4, 60, 120)

    def test_3instr(self):
        t1 = Track()
        t1.set_patch(3)
        t1.note(4, 93, 127)
        t2 = Track()
        t2.set_patch(4)
        t2.note(4, 94, 127)
        t3 = Track()
        t3.set_patch(5)
        t3.note(4, 95, 127)
        self.assertEqual(
            list(MidiEventStream(t1, t2, t3)),
            [('program-change', 0, 3),
             ('program-change', 1, 4),
             ('program-change', 2, 5),
             ('volume', 0, 100),
             ('note-on', 0, 93, 127),
             ('volume', 1, 100),
             ('note-on', 1, 94, 127),
             ('volume', 2, 100),
             ('note-on', 2, 95, 127),
             ('notelen-time', Rat(1, 4)),
             ('note-off', 0, 93, 127),
             ('note-off', 1, 94, 127),
             ('note-off', 2, 95, 127)])
        self.assertEqual(
            MidiEventStream(t1, t2, t3).str_repr(details=1),
            "p0:3 p1:4 p2:5 v0:100 n0:93 v1:100 n1:94 v2:100 n2:95 d1/4 o93 o94 o95")

    def test_str_repr(self):
        t1 = Track()
        t1.set_volume(88)
        t1.set_patch(3)
        t1.note(4, 93, 127)
        t2 = Track()
        t2.set_patch(4)
        t2.note(4, 94, 127)
        t3 = Track()
        t3.set_patch(5)
        t3.note(4, 95, 127)
        self.assertEqual(
            list(MidiEventStream(t1, t2, t3)),
            [('program-change', 0, 3),
             ('program-change', 1, 4),
             ('program-change', 2, 5),
             ('volume', 0, 88),
             ('note-on', 0, 93, 127),
             ('volume', 1, 100),
             ('note-on', 1, 94, 127),
             ('volume', 2, 100),
             ('note-on', 2, 95, 127),
             ('notelen-time', Rat(1, 4)),
             ('note-off', 0, 93, 127),
             ('note-off', 1, 94, 127),
             ('note-off', 2, 95, 127)])
        self.assertEqual(
            MidiEventStream(t1, t2, t3).str_repr(details=1),
            "p0:3 p1:4 p2:5 v0:88 n0:93 v1:100 n1:94 v2:100 n2:95 d1/4 o93 o94 o95")

    def test_track2(self):
        self.do_file("""
        header { random_transpose = no }
        question { music = music("\staff{ c'' }"
                                + "\\addvoice{ e' }"
                                + "\staff{ c }")
        }
        """)
        self.p._idx = 0
        mpd.music_to_tracklist(self.p.get_question()['music'].get_mpd_music_string(self.p))

    def test_track3(self):
        self.do_file("""
        header { random_transpose = no }
        question { music = music("\staff{ c''1 c''1 }"
                                + "\\addvoice{ r4 e'2. e'1 }"
                                + "\staff{ r4 r g2 g1 }"
                                + "\\addvoice{ r4 r r c c1}")
        }
        """)
        self.p._idx = 0
        tracklist = mpd.music_to_tracklist(self.p.get_question()['music'].get_mpd_music_string(self.p))
        track_fasit = ["n72 d1/1 o72 n72 d1/1 o72",
                       "d1/4 n64 d3/4 o64 n64 d1/1 o64",
                       "d1/2 n55 d1/2 o55 n55 d1/1 o55",
                       "d3/4 n48 d1/4 o48 n48 d1/1 o48"]
        for idx, correct in enumerate(track_fasit):
            self.assertEqual(tracklist[idx].str_repr(), correct)
        for idx in range(4):
            tracklist[idx].prepend_patch(idx + 1)
            track_fasit[idx] = "p%i " % (idx + 1) + track_fasit[idx]
        for idx, correct in enumerate(track_fasit):
            self.assertEqual(tracklist[idx].str_repr(), correct)
        self.assertEqual(
            MidiEventStream(*tracklist).str_repr(details=1),
            "p0:1 p1:2 p2:3 p3:4 v0:100 n0:72 d1/4 "
            "v1:100 n1:64 d1/4 "
            "v2:100 n2:55 d1/4 "
            "v3:100 n3:48 d1/4 "
            "o72 o64 o55 o48 "
            "n0:72 n1:64 n2:55 n3:48 d1/1 "
            "o72 o64 o55 o48")

    def test_track3_1(self):
        self.do_file("""
        header { random_transpose = no }
        question { music = music("\staff{ c''1 c''1 }"
                                + "\staff{ r4 r g2 g1 }")
        }
        """)
        self.p._idx = 0
        tracklist = mpd.music_to_tracklist(self.p.get_question()['music'].get_mpd_music_string(self.p))
        track_fasit = ["n72 d1/1 o72 n72 d1/1 o72",
                       "d1/2 n55 d1/2 o55 n55 d1/1 o55"]
        for idx, correct in enumerate(track_fasit):
            self.assertEqual(tracklist[idx].str_repr(), correct)
        for idx in range(2):
            tracklist[idx].prepend_patch(idx + 1)
            track_fasit[idx] = "p%i " % (idx + 1) + track_fasit[idx]
        for idx, correct in enumerate(track_fasit):
            self.assertEqual(tracklist[idx].str_repr(), correct)
        self.assertEqual(
            MidiEventStream(*tracklist).str_repr(details=1),
            "p0:1 p1:2 v0:100 n0:72 d1/2 v1:100 n1:55 d1/2 o72 o55 n0:72 n1:55 d1/1 o72 o55")

    def test_track3_2(self):
        self.do_file(r"""
        header { random_transpose = no }
        question { music = music("\staff{ c''1 c''1 }"
                                + "\staff{ r4 r g2 g1 }"
                                + "\staff{ r4 r r c c1}")
        }
        """)
        self.p._idx = 0
        tracklist = mpd.music_to_tracklist(self.p.get_question()['music'].get_mpd_music_string(self.p))
        track_fasit = ["n72 d1/1 o72 n72 d1/1 o72",
                       "d1/2 n55 d1/2 o55 n55 d1/1 o55",
                       "d3/4 n48 d1/4 o48 n48 d1/1 o48"]
        for idx, correct in enumerate(track_fasit):
            self.assertEqual(tracklist[idx].str_repr(), correct)
        for idx in range(3):
            tracklist[idx].prepend_patch(idx + 1)
            track_fasit[idx] = "p%i " % (idx + 1) + track_fasit[idx]
        for idx, correct in enumerate(track_fasit):
            self.assertEqual(tracklist[idx].str_repr(), correct)
        self.assertEqual(
            MidiEventStream(*tracklist).str_repr(details=1),
            "p0:1 p1:2 p2:3 v0:100 n0:72 d1/2 "
            "v1:100 n1:55 d1/4 "
            "v2:100 n2:48 d1/4 "
            "o72 o55 o48 "
            "n0:72 n1:55 n2:48 d1/1 "
            "o72 o55 o48")

    def test_bug1(self):
        """
        For each moment in time, all note-off events have to be
        done before the note-on events. This to avoid problems
        with the same note being played two times after each other
        in different tracks.
        """
        t1 = Track()
        t1.set_patch(3)
        t1.note(4, 93, 127)
        t1.note(4, 95, 127)
        t2 = Track()
        t2.set_patch(4)
        t2.note(4, 95, 127)
        t2.note(4, 97, 127)
        self.assertEqual(
            list(MidiEventStream(t1, t2)),
            [('program-change', 0, 3),
             ('program-change', 1, 4),
             ('volume', 0, 100),
             ('note-on', 0, 93, 127),
             ('volume', 1, 100),
             ('note-on', 1, 95, 127),
             ('notelen-time', Rat(1, 4)),
             ('note-off', 0, 93, 127),
             ('note-off', 1, 95, 127),
             ('note-on', 0, 95, 127),
             ('note-on', 1, 97, 127),
             ('notelen-time', Rat(1, 4)),
             ('note-off', 0, 95, 127),
             ('note-off', 1, 97, 127)])
        self.assertEqual(
            MidiEventStream(t1, t2).str_repr(details=1),
            "p0:3 p1:4 v0:100 n0:93 v1:100 n1:95 d1/4 o93 o95 n0:95 n1:97 d1/4 o95 o97")

    def test_melodic_interval_2_tracks(self):
        """
        In this test, only MIDI channel 0 will be allocated, even though
        two different patches and volumes are used. This because the tones
        from the two tracks does not sound at the same time.
        """
        t1 = Track()
        t1.set_patch(1)
        t1.set_volume(101)
        t1.note(4, 64)
        t2 = Track()
        t2.set_patch(2)
        t2.set_volume(102)
        t2.notelen_time(4)
        t2.note(4, 66)
        self.assertEqual(t1.str_repr(), "p1 v101 n64 d1/4 o64")
        self.assertEqual(t2.str_repr(), "p2 v102 d1/4 n66 d1/4 o66")
        m = MidiEventStream(t1, t2)
        self.assertEqual(
            m.str_repr(1),
            "p0:1 p1:2 v0:101 n0:64 d1/4 o64 v1:102 n1:66 d1/4 o66")

    def test_patch_volume_order(self):
        """
        Assert that the order of set_patch and set_volume does not matter.
        """
        t1 = Track()
        t1.set_patch(1)
        t1.set_volume(101)
        t1.note(4, 64)
        self.assertEqual(MidiEventStream(t1).str_repr(details=1), "p0:1 v0:101 n0:64 d1/4 o64")
        # Then with patch and volume in reverse order
        t1 = Track()
        t1.set_volume(101)
        t1.set_patch(1)
        t1.note(4, 64)
        self.assertEqual(MidiEventStream(t1).str_repr(details=1), "p0:1 v0:101 n0:64 d1/4 o64")

    def test_prepend_patch(self):
        """
        If multiple set_patch is done, the last will be used.
        """
        t = Track()
        t.prepend_patch(2)
        t.prepend_patch(3)
        t.note(4, 55)
        self.assertEqual(
            MidiEventStream(t).str_repr(), "p0:2 v0:100 n55 d1/4 o55")

    def test_set_patch(self):
        """
        If multiple set_patch is done, the last will be used.
        """
        t = Track()
        t.set_patch(2)
        t.set_patch(3)
        t.note(4, 55)
        self.assertEqual(
            MidiEventStream(t).str_repr(), "p0:3 v0:100 n55 d1/4 o55")

    def test_set_patch2(self):
        """
        Assert that there is not sendt a new volume event when we change patch.
        """
        t = Track()
        t.set_patch(2)
        t.note(4, 55)
        t.set_patch(3)
        t.note(4, 57)
        self.assertEqual(
            MidiEventStream(t).str_repr(),
            "p0:2 p1:3 v0:100 n55 d1/4 o55 v1:100 n57 d1/4 o57")

    def test_set_volume(self):
        """
        Assert that there is not sendt a new pacth event when we change volume
        """
        t = Track()
        t.set_volume(98)
        t.note(4, 55)
        t.set_volume(99)
        t.note(4, 57)
        self.assertEqual(
            MidiEventStream(t).str_repr(),
            "p0:0 p1:0 v0:98 n55 d1/4 o55 v1:99 n57 d1/4 o57")

    def test_set_bpm(self):
        t = Track()
        t.set_bpm(120)
        t.note(4, 50)
        self.assertEqual(
            MidiEventStream(t).str_repr(1),
            "t120/4 p0:0 v0:100 n0:50 d1/4 o50")

    def test_set_bpm2(self):
        """
        Two set_bpm in a row should only generate MIDI events for the
        last one.
        """
        t = Track()
        t.set_bpm(120)
        t.set_bpm(121)
        t.note(4, 50)
        self.assertEqual(
            MidiEventStream(t).str_repr(1),
            "t121/4 p0:0 v0:100 n0:50 d1/4 o50")

    def test_set_bpm3(self):
        """
        When two tracks set a different tempo, the tempo from the
        last track is used. There is not issues two MIDI events.
        """
        t1 = Track()
        t1.set_bpm(120)
        t1.note(4, 50)
        t2 = Track()
        t2.set_bpm(121)
        t2.note(4, 55)
        self.assertEqual(
            MidiEventStream(t1, t2).str_repr(1),
            "t121/4 p0:0 v0:100 n0:50 n0:55 d1/4 o50 o55")

    def test_x1(self):
        t1 = Track()
        t1.set_patch(1)
        t1.note(4, 60)
        t1.set_patch(2)
        t1.note(4, 60)
        t1.set_patch(3)
        t1.note(4, 60)
        t1.set_patch(4)
        t1.note(4, 60)
        m = MidiEventStream(t1)
        self.assertEqual("p0:1 p1:2 p2:3 p3:4 v0:100 n0:60 d1/4 o60 v1:100 n1:60 d1/4 o60 v2:100 n2:60 d1/4 o60 v3:100 n3:60 d1/4 o60", m.str_repr(1))

    def test_x2(self):
        t1 = Track()
        t1.note(4, 62)
        t1.note(4, 60)
        t1.note(2, 62)
        t2 = Track()
        t2.note(1, 60)
        m = MidiEventStream(t1, t2)
        self.assertEqual("p0:0 p1:0 v0:100 n0:62 n0:60 d1/4 o62 v1:100 n1:60 d1/4 o60 n0:62 d1/2 o62 o60", m.str_repr(1))

    def test_x3(self):
        t1 = Track()
        t1.note(4, 62)
        t1.note(4, 60)
        t1.note(2, 62)
        t2 = Track()
        t2.note(1, 60)
        m = MidiEventStream(t1, t2)
        m.num_MIDI_channels = 1
        # We don't handle running out of MIDI channels yet
        self.assertRaises(Exception, m.str_repr, 1)


class TestChannelDevice(unittest.TestCase):

    def setUp(self):
        self.cd = MidiEventStream.ChannelDevice()

    def test_1(self):
        ch_dev = MidiEventStream.ChannelDevice()
        self.assertEqual(0, ch_dev.require_channel(60, 0, 100))
        ch_dev.start_note(0, 60)
        self.assertEqual(1, ch_dev.require_channel(61, 1, 100))
        ch_dev.start_note(1, 61)
        self.assertEqual(2, ch_dev.require_channel(62, 1, 70))
        ch_dev.start_note(2, 62)
        self.assertEqual(0, ch_dev.require_channel(63, 0, 100))
        ch_dev.start_note(0, 63)
        # New channel since we want to play another tone with the
        # same pitch as the one playing in channel 0
        self.assertEqual(3, ch_dev.require_channel(63, 0, 100))
        ch_dev.start_note(3, 63)
        ch_dev.stop_note(0, 60)
        ch_dev.stop_note(1, 61)
        ch_dev.stop_note(2, 62)
        ch_dev.stop_note(0, 63)
        self.assertEqual(0, ch_dev.require_channel(63, 0, 100))
        ch_dev.start_note(0, 63)

suite = unittest.makeSuite(TestTrack)
suite.addTest(unittest.makeSuite(TestMidiEventStream))
suite.addTest(unittest.makeSuite(TestChannelDevice))



#!/usr/bin/python2.3

import sys, os
if os.getcwdu()[-4:] == "test":
    sys.path.insert(0, "..")
else:
    sys.path.insert(0, ".")
import soundcard
soundcard.initialise_devicefile("/dev/sequencer2", 2)

s = soundcard.solfege_c_midi

devnum = soundcard.synth.m_devnum
chan = 0
s.seq_start_timer()
#s.seq_set_patch(devnum, chan, 48)

# these two lines are necessary to make the first tone stop.
#s.seq_start_note(devnum, chan, 60, 90)
#s.seq_stop_note(devnum, chan, 60, 90)

# for i in range(500):
#        s.seq_start_note(devnum, chan, 60 + i % 10, 90)
#        s.seq_delta_time(25)
#        s.seq_stop_note(devnum, chan, 60 + i % 10, 90)
s.seq_set_patch(0, 0, 63)
s.seq_set_patch(0, 1, 63)
for i in range(500):
        s.seq_start_note(devnum, 0, 60 + (i * 2) % 10, 90)
        s.seq_delta_time(25)
        s.seq_stop_note(devnum, 0, 60 + (i * 2) % 10, 90)
        s.seq_start_note(devnum, 1, 60 + (i * 2) % 10 + 1, 90)
        s.seq_delta_time(25)
        s.seq_stop_note(devnum, 1, 60 + (i * 2) % 10 + 1, 90)


s.seqbuf_dump()

import sys
print "press ENTER to continue"
sys.stdin.readline()
soundcard.synth.close()



# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import random
import os

from gi.repository import GObject
from gi.repository import Gtk

from solfege import abstract
from solfege import gu
from solfege import inputwidgets
from solfege import lessonfile
from solfege import mpd
from solfege import soundcard
from solfege import statistics, statisticsviewer
from solfege import utils

import solfege


class Teacher(abstract.Teacher):
    # FIXME the following lines
    OK, ERR_PICKY, ERR_TONES = list(range(3))
    ERR_PICKY = 1
    ERR_CONFIG = 2
    OCTAVES = [-2, -1, 0, 1, 2, 3]

    def __init__(self, exname):
        abstract.Teacher.__init__(self, exname)
        self.lessonfileclass = lessonfile.HeaderLessonfile
        self.m_statistics = statistics.IdToneStatistics(self)
        self.m_ask_tones = {}
        self.m_question = None
        self.m_custom_mode = False

    def new_question(self):
        """
        Return values:
        OK: sucess, new random tone selected
        ERR_PICKY: fail, you are not allowed to select a new tone before you
                   can identify the one you have now.
        ERR_CONFIG: fail, all notes have zero weight or no octaves selected
        """
        if self.m_timeout_handle:
            GObject.source_remove(self.m_timeout_handle)
            self.m_timeout_handle = None

        if self.get_bool('config/picky_on_new_question') \
                and self.q_status in (self.QSTATUS_NEW, self.QSTATUS_WRONG):
            return Teacher.ERR_PICKY

        self.m_is_first_question = self.q_status == self.QSTATUS_NO

        v = []
        for n in mpd.MusicalPitch.notenames:
            v.extend([n] * self.get_int(n + "_weight"))
        if not v:
            return self.ERR_CONFIG
        self.m_question = random.choice(v)
        v = []
        for n in self.OCTAVES:
            if self.get_bool("octave" + str(n)):
                v.append(n)
        if not v:
            return self.ERR_CONFIG
        self.m_octave = random.choice(v)
        self.q_status = self.QSTATUS_NEW
        return self.OK

    def guess_answer(self, notename):
        if notename == self.m_question:
            if self.q_status == self.QSTATUS_NEW:
                self.m_statistics.add_correct(notename)
            self.maybe_auto_new_question()
            self.q_status = self.QSTATUS_SOLVED
            return 1
        else:
            if self.q_status == self.QSTATUS_NEW:
                self.m_statistics.add_wrong(self.m_question, notename)
                self.q_status = self.QSTATUS_WRONG

    def play_question(self):
        if self.q_status == self.QSTATUS_NO:
            return
        utils.play_note(4,
            mpd.notename_to_int(self.m_question) + self.m_octave * 12)

    def give_up(self):
        self.q_status = self.QSTATUS_GIVE_UP

    def spank_me_play_question(self):
        t1 = utils.new_percussion_track()
        t1.note(8, 71)
        t2 = utils.new_track()
        t2.notelen_time(4)
        t2.note(4, mpd.notename_to_int(self.m_question) + self.m_octave * 12)
        soundcard.synth.play_track(t1, t2)

    def spank_me(self):
        utils.play_perc(4, 71)


class Gui(abstract.Gui):

    def __init__(self, teacher):
        abstract.Gui.__init__(self, teacher)

        self.g_percentage = gu.bLabel(self.practise_box, "")
        self.g_percentage.set_name("Heading1")
        self.g_piano = inputwidgets.PianoOctaveWithAccelName(
                       self.on_answer_from_user, self.get_accel_key_list())
        self.g_piano.m_visible_accels = not self.get_bool('hide_piano_accels')

        def update_accels(*ignore):
            self.g_piano.m_keys = self.get_accel_key_list()
            self.g_piano.queue_draw()
        for notename in mpd.MusicalPitch.notenames:
            self.add_watch('tone_%s_ak' % notename, update_accels)
        self.practise_box.pack_start(self.g_piano, True, True, 0)

        self.g_flashbar = gu.FlashBar()
        self.g_flashbar.show()
        self.practise_box.pack_start(self.g_flashbar, False, False, 0)
        self.practise_box.set_spacing(gu.PAD)

        self.std_buttons_add(
            ('new-tone', self.new_question),
            ('repeat', lambda _o, self=self: self.m_t.play_question()),
            ('give_up', self.give_up))
        self.practise_box.show_all()
        ###############
        # config_grid #
        ###############
        table = Gtk.Table()
        table.set_border_width(gu.PAD_SMALL)
        self.g_tones_frame = frame = Gtk.Frame(label=_("Weight"))
        self.g_config_grid.attach(frame, 0, 0, 3, 1)
        frame.add(table)
        for x, n in [(1, 'cis'), (3, 'dis'), (7, 'fis'),
                     (9, 'gis'), (11, 'ais')]:
            label = Gtk.Label(label=mpd.MusicalPitch.new_from_notename(n).get_user_notename())
            label.set_name("Heading2")
            label.set_alignment(0.2, 1.0)
            table.attach(label, x, x + 2, 0, 1, xoptions=Gtk.AttachOptions.FILL)
            b = gu.nSpinButton(self.m_exname, n + "_weight",
                      Gtk.Adjustment(1, 0, 1000, 1, 10), digits=0)
            table.attach(b, x, x + 2, 1, 2, xoptions=Gtk.AttachOptions.FILL)
        for x, n in [(0, 'c'), (2, 'd'), (4, 'e'), (6, 'f'),
                      (8, 'g'), (10, 'a'), (12, 'b')]:
            label = Gtk.Label(label=mpd.MusicalPitch.new_from_notename(n).get_user_notename())
            label.set_name("Heading2")
            label.set_alignment(0.35, 1.0)
            table.attach(label, x, x + 2, 2, 3, xoptions=Gtk.AttachOptions.FILL)
            b = gu.nSpinButton(self.m_exname, n + "_weight",
                   Gtk.Adjustment(1, 0, 1000, 1, 10), digits=0)
            table.attach(b, x, x + 2, 3, 4, xoptions=Gtk.AttachOptions.FILL)

        self.g_octaves_select = hbox = Gtk.HBox()
        self.g_config_grid.attach(hbox, 0, 1, 3, 1)
        hbox.pack_start(Gtk.Label(_("Octave:")), False, False, padding=4)
        for oct in self.m_t.OCTAVES:
            b = gu.nCheckButton(self.m_exname, "octave" + str(oct), str(oct),
                                default_value=1)
            hbox.pack_start(b, False, False, 0)
        #############
        self._add_auto_new_question_gui(row=4)
        #############
        b = gu.nCheckButton('idtone', 'hide_piano_accels', _("Hide _piano keyboard shortcuts"), False)

        def show_hide_accels(checkbutton):
            self.g_piano.m_visible_accels = not b.get_active()
        b.connect('clicked', show_hide_accels)
        self.g_config_grid.attach(b, 0, 2, 1, 1)
        #
        self.g_config_grid.attach(gu.nCheckButton(self.m_exname,
                    "warning_sound",
                    _("Play warning sound when you guess wrong")),
                0, 3, 1, 1)
        self.g_config_grid.show_all()
        ##############
        # statistics #
        ##############
        self.setup_statisticsviewer(statisticsviewer.StatisticsViewer,
                                   _("Identify tone"))

    def get_accel_key_list(self):
        v = []
        for k in mpd.MusicalPitch.notenames:
            self.m_key_bindings['tone_%s_ak' % k] \
                = lambda self=self, k=k: self.on_answer_from_user(k)
            v.append(self.get_string('tone_%s_ak' % k))
        return v

    def new_question(self, widget=None):
        g = self.m_t.new_question()
        if g == Teacher.ERR_CONFIG:
            solfege.win.display_error_message(
_("""You have to select some tones practise. Do this on the config page by setting the weight of tones to a value greater than zero."""))
            return
        elif g == Teacher.OK:
            self.std_buttons_new_question()
            try:
                if self.m_t.m_is_first_question:
                    self.flash_and_play_first_tone()
                else:
                    self.g_flashbar.clear()
                    self.m_t.play_question()
            except Exception as e:
                def cleanup():
                    self.std_buttons_exception_cleanup()
                if not self.standard_exception_handler(e, cleanup):
                    raise
        self.set_percentage_label()

    def flash_and_play_first_tone(self):
        self.g_flashbar.flash(_("First tone is %s") % mpd.MusicalPitch.new_from_notename(self.m_t.m_question).get_user_notename())
        self.m_t.play_question()

    def on_answer_from_user(self, notename):
        if self.m_t.q_status == self.QSTATUS_NO:
            self.g_flashbar.flash(_("Click 'New tone' to begin."))
            return
        elif self.m_t.q_status == self.QSTATUS_SOLVED:
            if self.m_t.guess_answer(notename):
                self.g_flashbar.flash(_("Correct, but you have already solved this question"))
            else:
                self.g_flashbar.flash(_("Wrong, but you have already solved this question"))
        elif self.m_t.q_status in (self.QSTATUS_NEW, self.QSTATUS_WRONG):
            if self.m_t.guess_answer(notename):
                self.g_flashbar.flash(_("Correct"))
                self.std_buttons_answer_correct()
            else:
                try:
                    if self.m_t.m_is_first_question:
                        self.flash_and_play_first_tone()
                        return
                    self.g_flashbar.flash(_("Wrong"))
                    self.std_buttons_answer_wrong()
                    if self.get_bool("warning_sound"):
                        if self.get_bool("config/auto_repeat_question_if_wrong_answer"):
                            self.m_t.spank_me_play_question()
                        else:
                            self.m_t.spank_me()
                    else:
                        if self.get_bool("config/auto_repeat_question_if_wrong_answer"):
                            self.m_t.play_question()
                except Exception as e:
                    if not self.standard_exception_handler(e):
                        raise
        self.set_percentage_label()

    def give_up(self, _o=None):
        if self.m_t.q_status == self.QSTATUS_WRONG:
            self.g_flashbar.push(_("The answer is: %s")
                % mpd.MusicalPitch.new_from_notename(self.m_t.m_question).get_user_notename())
            self.m_t.give_up()
            self.std_buttons_give_up()

    def set_percentage_label(self):
        self.g_percentage.set_text("%.1f %%" % (self.m_t.m_statistics.get_percentage_correct()))

    def on_start_practise(self):
        self.m_t.m_custom_mode = not (
                ('white_keys_weight' in self.m_t.m_P.header)
                 or ('black_keys_weight' in self.m_t.m_P.header))
        super(Gui, self).on_start_practise()
        self.g_flashbar.require_size([
            _("Click 'New tone' to begin."),
            _("Correct, but you have already solved this question"),
            _("Wrong, but you have already solved this question"),
        ])
        if self.m_t.m_custom_mode:
            for notename, value in zip(mpd.MusicalPitch.notenames,
                                       self.get_list('custom_mode_cfg')):
                try:
                    value = float(value)
                except ValueError:
                    value = 0.0
                self.set_float('%s_weight' % notename, value)
        else:
            if 'white_keys_weight' in self.m_t.m_P.header:
                if type(self.m_t.m_P.header.white_keys_weight) == list \
                        and len(self.m_t.m_P.header.white_keys_weight) == 7:
                    for idx, n in enumerate(mpd.MusicalPitch.natural_notenames):
                        try:
                            weight = float(self.m_t.m_P.header.white_keys_weight[idx])
                        except ValueError:
                            weight = 0.0
                        self.set_float('%s_weight' % n, weight)
                else:
                    gu.dialog_ok("The white_keys_weight variable in the lesson file '%s' had wrong type" % os.path.abspath(self.m_t.m_P.m_filename), self, msgtype=Gtk.MessageType.WARNING)
            else:
                for idx, n in enumerate(mpd.MusicalPitch.notenames):
                    self.set_float('%s_weight' % n, 0.0)
            if 'black_keys_weight' in self.m_t.m_P.header:
                if type(self.m_t.m_P.header.black_keys_weight) == list \
                        and len(self.m_t.m_P.header.black_keys_weight) == 5:
                    for idx, n in enumerate(mpd.MusicalPitch.sharp_notenames):
                        try:
                            weight = float(self.m_t.m_P.header.black_keys_weight[idx])
                        except ValueError:
                            weight = 0.0
                        self.set_float('%s_weight' % n, weight)
                else:
                    gu.dialog_ok("The black_keys_weight variable in the lesson file '%s' had wrong type" % os.path.abspath(self.m_t.m_P.m_filename), self, msgtype=Gtk.MessageType.WARNING)
            else:
                for idx, n in enumerate(('cis', 'dis', 'fis', 'gis', 'ais')):
                    self.set_float('%s_weight' % n, 0.0)
        if self.m_t.m_custom_mode:
            self.g_tones_frame.show()
            self.g_octaves_select.show()
            self.m_t.m_statistics.reset_custom_mode_session(self.m_t.m_P.m_filename)
        else:
            self.g_tones_frame.hide()
            self.g_octaves_select.hide()
            self.m_t.m_statistics.reset_session()
        self.g_statview.g_heading.set_text("%s - %s" % (_("Identify tone"), self.m_t.m_P.header.title))
        self.set_percentage_label()
        self.g_flashbar.delayed_flash(self.short_delay,
            _("Click 'New tone' to begin."))
        self.std_buttons_start_practise()
        self.m_t.q_status = self.QSTATUS_NO

    def on_end_practise(self):
        if self.m_t.m_custom_mode:
            self.set_list('custom_mode_cfg', [self.get_float('%s_weight' % x)
                         for x in mpd.MusicalPitch.notenames])
        self.m_t.end_practise()
        self.std_buttons_end_practise()



# vim: set fileencoding=utf-8:
# GNU Solfege - free ear training software
# Copyright (C) 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
import random
import time
import inspect

from gi.repository import Gtk

import solfege
from solfege import abstract
from solfege import cfg
from solfege import gu
from solfege import lessonfile
from solfege import mpd
from solfege import soundcard
from solfege import statistics
from solfege import statisticsviewer
from solfege.mpd import elems

labels = ['1', '♯1/♭2', '2', '♯2/♭3', '3', '4', '♯4/♭5',
          '5', '♯5/6♭', '6', '♯6/♭7', '7', '1']


class ToneInKeyStatistics(statistics.LessonStatistics):

    def add_correct(self, answer):
        fileid = solfege.db.get_fileid(self.m_t.m_P.m_filename)
        cursor = solfege.db.conn.cursor()
        cursor.execute(
            "insert into toneincontext "
            "(fileid, timestamp, answerkey, guessedkey) "
            "values (?, ?, ?, ?)",
            (fileid, int(time.time()), answer, answer))
        solfege.db.conn.commit()

    def add_wrong(self, question, answer):
        if question == 12:
            question = 0
        if answer == 12:
            answer = 0
        fileid = solfege.db.get_fileid(self.m_t.m_P.m_filename)
        cursor = solfege.db.conn.cursor()
        cursor.execute(
            "insert into toneincontext "
            "(fileid, timestamp, answerkey, guessedkey) "
            "values (?, ?, ?, ?)",
            (fileid, int(time.time()), question, answer))
        solfege.db.conn.commit()

    def key_to_pretty_name(self, key):
        return labels[int(key)]

    def num_asked(self, key):
        row = solfege.db.conn.execute(
            "select count(guessedkey) from toneincontext "
            "where answerkey=? and fileid=?",
            (key, solfege.db.get_fileid(self.m_t.m_P.m_filename)))
        for r in row:
            return r[0]
        return 0

    def num_correct(self, key):
        row = solfege.db.conn.execute(
            "select count(guessedkey) from toneincontext "
            "where answerkey=? and guessedkey=? and fileid=?",
            (key, key, solfege.db.get_fileid(self.m_t.m_P.m_filename)))
        for r in row:
            return r[0]
        return 0

    @staticmethod
    def get_percentage_of_file(filename):
        """
        raise solfege.statistics.DB.FileNotInDB if the lesson file never
        has been practised.
        """
        # how many answered correct
        row = solfege.db.conn.execute(
            "select count(guessedkey) from toneincontext "
            "where answerkey=guessedkey and fileid=?",
            (solfege.db.get_fileid(filename),))
        correct = row.fetchone()
        correct = correct[0] if correct else 0
        # how many asked
        row = solfege.db.conn.execute(
            "select count(guessedkey) from toneincontext "
            "where fileid=?",
            (solfege.db.get_fileid(filename),))
        asked = row.fetchone()
        asked = asked[0] if correct else 1
        return correct / asked


class StatisticsViewer(Gtk.Grid):

    def __init__(self, teacher):
        Gtk.Grid.__init__(self)
        self.set_column_spacing(gu.hig.SPACE_SMALL)
        self.set_row_spacing(gu.hig.SPACE_SMALL)
        self.props.margin = gu.hig.SPACE_SMALL
        self.m_t = teacher
        self.set_column_homogeneous(True)
        self.labels = fill_grid(Gtk.Label, self)
        self.show_all()

    def update(self, *w):
        for n in range(13):
            c = self.m_t.m_statistics.num_correct(n if n < 12 else 0)
            a = self.m_t.m_statistics.num_asked(n if n < 12 else 0)
            p = 0 if a == 0 else c / a
            if n not in self.m_t.get_list("tones"):
                self.labels[n].set_name("labelgreyout")
            elif p < 0.5:
                self.labels[n].set_name("labelred")
            elif p < 0.9:
                self.labels[n].set_name("labelyellow")
            else:
                self.labels[n].set_name("labelgreen")
            self.labels[n].set_text("{}%\n{}/{}".format(int(p * 100), c, a))


class Teacher(abstract.Teacher):
    OK = 0
    ERR_NO_TONES = 1
    resolve = {
        'major': {
            0: "c2",
            1: "cis c2",
            2: "d c2",
            3: "dis4 d8 c2",
            4: "e4 d8 c2",
            5: "f4 e8 d c2",
            6: "fis4 g8 a b c'2",
            7: "g4 a8 b c'2",
            8: "gis4 a8 b c'2",
            9: "a4 b8 c'2",
            10: "ais4 b8 c'2",
            11: "b4 c'2",
            12: "c'2"},
        'minor': {
            0: "c2",
            1: "des c2",
            2: "d c2",
            3: "es4 d8 c2",
            4: "e4 es8 d c2",
            5: "f4 es8 d c2",
            6: "fis4 g8 as bes c'2",
            7: "g4 as8 bes c'2",
            8: "as4 bes8 c'2",
            9: "a4 bes8 c'2",
            10: "bes4 c'2",
            11: "b4 c'2",
            12: "c'2"}
    }
    cadence = {
        'major': {
            'music': lessonfile.Music(r"\staff\relative g'{ \time 3/4 <g e c> <a f c> <g f d b> <g2 e c> }"),
            'name': _("Major"),
            'key': 'major',
        },
        'minor': {
            'music': lessonfile.Music(r"\staff\relative g'{ \time 3/4 <g es c> <as f c> <g f d b> <g2 es c> }"),
            'name': _("Minor"),
            'key': 'minor',
        }
    }

    def __init__(self, exname):
        abstract.Teacher.__init__(self, exname)
        self.lessonfileclass = lessonfile.HeaderLessonfile
        self.m_lessonfile_header_defaults['random_tonic'] = False
        self.m_lessonfile_header_defaults['cadence'] = 'major'
        self.m_lessonfile_defs['other'] = 'other'
        self.m_transpose = mpd.MusicalPitch()
        """To which key the question should be transposed.
        Will be set in :func:`~new_question`
        """
        self.m_statistics = ToneInKeyStatistics(self)

    def new_question(self):
        """
        Return OK or ERR_NO_TONES
        Will set the following variables defining the question.

            self.m_tone
            self.m_octave
            self.m_transpose
        """
        if not self.get_list("tones"):
            return self.ERR_NO_TONES
        self.m_tone = random.choice(
            [i for i in self.get_list("tones") if i != 12])
        if self.get_bool('many_octaves'):
            self.m_octave = random.choice((-1, 0, 1, 2))
        else:
            self.m_octave = 0
        self.q_status = self.QSTATUS_NEW
        if self.get_string('random_tonic') == 'True':
            self.m_transpose.randomize("c'", "b'")
        if self.m_custom_mode:
            # In m_custom_mode, we will choose one of the cadences selected
            # on the config page.
            cadence_list = [k for k in list(self.m_cadences.keys()) if self.m_cadences[k]]
            if not cadence_list:
                return self.ERR_NO_CADENCES
            self.m_chosen_cadence = c = random.choice(cadence_list)
            if isinstance(c, str):
                self.m_cadence = self.cadence[c]
            else:
                self.m_cadence = self.m_P.blocklists['cadence'][c]
        else:
            # In normal mode (not m_custom_mode), if there are cadences
            # defined in the lesson file, we choose one of them by random.
            if 'cadence' in self.m_P.blocklists:
                self.m_cadence = random.choice(self.m_P.blocklists['cadence'])
            else:
                # If no cadences are defined in the lesson file, the header
                # variable cadence should select one of the builtin cadences.
                self.m_cadence = self.cadence[self.m_P.header.cadence]
        return self.OK

    def play_answer(self):
        soundcard.synth.play_track(self.get_track_of_answer())

    def play_question(self):
        """
        Play the resolution of the previous question, if any, and then
        the cadence and the new question.
        """
        t = self.get_track_of_question()
        soundcard.synth.play_track(t)

    def guess_answer(self, answer):
        if answer == 12:
            answer = 0
        assert self.q_status not in (self.QSTATUS_NO, self.QSTATUS_GIVE_UP)
        if self.m_tone == answer or (answer == 0 and self.m_tone == 12):
            if self.q_status == self.QSTATUS_NEW:
                self.m_statistics.add_correct(answer)
            answer_track = self.get_track_of_answer()
            self.new_question()
            question_track = self.get_track_of_question()
            soundcard.synth.play_track(answer_track + question_track)
            self.q_status = self.QSTATUS_NEW
            return 1
        else:
            if self.q_status == self.QSTATUS_NEW:
                self.m_statistics.add_wrong(self.m_tone, answer)
                self.q_status = self.QSTATUS_WRONG

    def get_track_of_answer(self):
        """
        If the lessonfile have defined a resolve block with the same key
        as the cadense that is selected, it will use that. If not it will
        use the builtin.
        """
        # First see if we have a matching resolve block in the lesson file
        try:
            resolve = [r for r in self.m_P.blocklists['resolve']
                       if r['key'] == self.m_cadence['key']][0]['list'][self.m_tone]
        except (IndexError, KeyError):
            # If not found, we use the builtin definition
            resolve = self.resolve[self.m_cadence['key']][self.m_tone]

        tr = self.m_transpose.clone()
        tr.m_octave_i += self.m_octave - 1
        t = mpd.music_to_track(r"\staff\transpose %s'{ \time 100/4 %s }" % (
            tr.get_octave_notename(), resolve))
        t.prepend_bpm(90)
        return t

    def get_track_of_question(self):
        c = self.m_cadence['music'].get_mpd_music_string(self.m_P)
        c = c.replace(r"\transpose c'",
                      r"\transpose %s" % self.m_transpose.get_octave_notename())
        cadence_track = mpd.music_to_track(c)
        cadence_track.prepend_bpm(90)

        p = mpd.MusicalPitch.new_from_notename("c'") + self.m_tone
        p.m_octave_i = self.m_octave
        s = r"\staff{ %s }" % p.transpose_by_musicalpitch(self.m_transpose).get_octave_notename()
        track = mpd.music_to_track(s)
        return cadence_track + track

    def give_up(self):
        self.m_qstatus = self.QSTATUS_GIVE_UP

    def start_practise(self):
        self.m_custom_mode = bool(not self.m_P.header.tones)
        if self.m_custom_mode:
            self.m_statistics.reset_custom_mode_session(self.m_P.m_filename)
        else:
            self.m_statistics.reset_session()
        if self.m_P.header.tones:
            self.set_list("tones", self.m_P.header.tones)
        # FIXME in preparation for 'configure yourself mode'
        self.set_string('random_tonic', str(self.m_P.header.random_tonic))
        if self.get_string('random_tonic') == 'other':
            self.m_transpose.randomize("cis'", "b'")
        elif self.get_string('random_tonic') == 'False':
            self.m_transpose.set_from_notename("c'")


def fill_grid(button_class, grid):
    grid.set_column_homogeneous(True)
    buttons = {}
    for p, x in ((0, 1), (1, 3), (3, 6), (4, 8), (5, 10)):
        b = button_class(labels[x])
        buttons[x] = b
        grid.attach(b, p * 2 + 1, 0, 2, 1)
    for p, x in enumerate((0, 2, 4, 5, 7, 9, 11, 12)):
        b = button_class(labels[x])
        buttons[x] = b
        grid.attach(b, p * 2, 1, 2, 1)
    return buttons


class nConfigButtons(Gtk.Grid, cfg.ConfigUtils):

    def __init__(self, exname, name):
        Gtk.Grid.__init__(self)
        cfg.ConfigUtils.__init__(self, exname)
        self.m_varname = name
        self.g_buttons = fill_grid(Gtk.CheckButton, self)
        for key, button in list(self.g_buttons.items()):
            button.connect('toggled', self.on_toggled)
        for key in self.get_list('tones'):
            self.g_buttons[key].set_active(True)

    def on_toggled(self, *w):
        self.set_list(self.m_varname,
            [k for k in list(self.g_buttons.keys()) if self.g_buttons[k].get_active()])


class Gui(abstract.Gui):
    lesson_heading = _("Tone in context")

    def __init__(self, teacher):
        abstract.Gui.__init__(self, teacher)
        grid = Gtk.Grid()
        grid.set_row_spacing(gu.hig.SPACE_SMALL)
        grid.set_column_spacing(gu.hig.SPACE_SMALL)
        self.g_buttons = fill_grid(Gtk.Button, grid)
        for key, button in list(self.g_buttons.items()):
            button.connect('clicked', self.on_left_click, key)
        self.practise_box.pack_start(grid, False, False, gu.hig.SPACE_SMALL)
        self.g_flashbar = gu.FlashBar()
        self.g_flashbar.show()
        self.practise_box.pack_start(self.g_flashbar, False, False, 0)
        self.std_buttons_add(
            ('new', self.new_question),
            ('repeat', lambda _o, self=self: self.m_t.play_question()),
            ('give_up', self.give_up))
        self.practise_box.show_all()
        ###############
        # config_grid #
        ###############
        # FIXME we need a ComboBox
        # self.g_random = gu.nCheckButton(self.m_exname, 'random_tonic', _("Random transpose"))
        # self.g_config_grid.attach(self.g_random, 0, 0, 1, 1)
        #
        self.g_tones_category, box = gu.hig_category_vbox(_("Tones"))
        self.g_config_grid.attach(self.g_tones_category, 0, 2, 3, 1)
        self.g_tone_selector = nConfigButtons(self.m_exname, 'tones')
        self.g_tone_selector.show_all()
        box.pack_start(self.g_tone_selector, False, False, 0)
        self.g_many_octaves = b = gu.nCheckButton('toneincontext', 'many_octaves',
            _("Many octaves"))
        self.g_config_grid.attach(b, 0, 3, 1, 1)
        # Cadences
        self.g_cadences_category, self.g_cadences = gu.hig_category_vbox(_("Cadences"))
        self.g_cadences.show()
        self.g_config_grid.attach(self.g_cadences_category, 0, 4, 1, 1)
        #

        def _ff(var):
            if self.m_t.m_custom_mode:
                # If we are running in custom mode, then the user can
                # select himself what intervals to practise. And then
                # we have to reset the exercise.
                # self.on_end_practise()
                # self.on_start_practise()
                self.cancel_question()
        self.add_watch('tones', _ff)
        ##############
        # Statistics #
        ##############
        self.setup_statisticsviewer(statisticsviewer.StatisticsViewer,
                                   _("Tone in cadence"))
        #
        self.g_statview = StatisticsViewer(self.m_t)
        self.g_statview.show()
        self.g_notebook.append_page(self.g_statview, Gtk.Label(label=_("Statistics")))
        self.g_notebook.connect(
            'switch_page', self.g_statview.update)

    def cancel_question(self):
        self.m_t.end_practise()
        self.std_buttons_end_practise()

    def new_question(self, *w):
        i = self.m_t.new_question()
        if i == Teacher.OK:
            self.std_buttons_new_question()
            self.m_t.play_question()
            for key, button in list(self.g_buttons.items()):
                button.set_sensitive(key in self.get_list("tones"))
        elif i == Teacher.ERR_NO_CADENCES:
            self.g_flashbar.flash(_("No cadences selected"))

    def give_up(self, *w):
        if self.m_t.q_status == self.QSTATUS_WRONG:
            self.g_flashbar.push(_("The answer is: %s")
                 % labels[self.m_t.m_tone])
            self.m_t.give_up()
            self.std_buttons_give_up()

    def on_left_click(self, button, tone_int):
        if self.m_t.q_status == self.QSTATUS_SOLVED:
            if self.m_t.guess_answer(tone_int):
                self.g_flashbar.flash(_("Correct, but you have already solved this question"))
            else:
                self.g_flashbar.flash(_("Wrong, but you have already solved this question"))
        elif self.m_t.q_status in (self.QSTATUS_NEW, self.QSTATUS_WRONG):
            if self.m_t.guess_answer(tone_int):
                self.g_flashbar.flash(_("Correct"))
                self.std_buttons_new_question()
            else:
                self.g_flashbar.flash(("Wrong"))
                self.std_buttons_answer_wrong()
                if self.get_bool("config/auto_repeat_question_if_wrong_answer"):
                    self.m_t.play_question()

    def on_start_practise(self):
        self.m_t.start_practise()
        super(Gui, self).on_start_practise()
        if self.m_t.m_custom_mode:
            self.g_tone_selector.show()
            # self.g_random.show()
            self.g_tones_category.show()
            self.g_many_octaves.show()
            for w in self.g_cadences.get_children():
                w.destroy()
            # We let the user select between all available cadences, both
            # the builtin cadences and any defined in the lesson file.
            # First we add gui for the builtin cadences
            self.g_cadences_category.show()
            self.g_cadences.show()
            self.m_t.m_cadences = {}
            self.m_t.m_cadences['minor'] = False
            self.m_t.m_cadences['major'] = False
            b = _(" (builtin)") if 'cadence' in self.m_t.m_P.blocklists \
               else ""
            for k in Teacher.cadence:
                btn = Gtk.CheckButton(Teacher.cadence[k]['name'] + b)
                btn.show()
                btn.set_active(True)
                self.m_t.m_cadences[k] = True
                btn.connect('toggled', self.on_cadences_toggled, k)
                self.g_cadences.pack_start(btn, False, False, 0)
            # Then for the one supplied with the lesson file
            if 'cadence' in self.m_t.m_P.blocklists:
                for idx, c in enumerate(self.m_t.m_P.blocklists['cadence']):
                    name = c.get('name', _("Unnamed"))
                    btn = Gtk.CheckButton(name)
                    btn.show()
                    btn.set_active(True)
                    self.m_t.m_cadences[idx] = True
                    btn.connect('toggled', self.on_cadences_toggled, idx)
                    self.g_cadences.pack_start(btn, False, False, 0)
        else:
            self.g_tone_selector.hide()
            self.g_tones_category.hide()
            self.g_many_octaves.hide()
            self.g_cadences_category.hide()
            # self.g_random.hide()
        for key, button in list(self.g_buttons.items()):
            button.set_sensitive(False)
        self.set_bool('tone_in_cadence', self.m_t.m_P.header.tone_in_cadence)
        self.std_buttons_start_practise()
        self.g_flashbar.delayed_flash(self.short_delay,
            _("Click 'New' to begin."))
        self.g_flashbar.require_size([
            _("Correct, but you have already solved this question"),
            _("Wrong, but you have already solved this question")])

    def on_cadences_toggled(self, btn, key):
        self.cancel_question()
        self.m_t.m_cadences[key] = btn.get_active()



# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from solfege import abstract
from solfege import gu
from solfege import lessonfile
from solfege import mpd
from solfege import soundcard
from solfege import utils


class Teacher(abstract.Teacher):

    def __init__(self, exname):
        abstract.Teacher.__init__(self, exname)
        self.lessonfileclass = lessonfile.SingChordLessonfile
        for s in 'accidentals', 'key', 'semitones', 'atonal':
            self.m_lessonfile_defs[s] = s

    def new_question(self):
        self.q_status = self.QSTATUS_NEW
        self.m_P.select_random_question()

    def play_440hz(self):
        utils.play_note(4, mpd.notename_to_int("a'"))


class Gui(abstract.LessonbasedGui):

    def __init__(self, teacher):
        abstract.LessonbasedGui.__init__(self, teacher)
        ################
        # practise_box #
        ################
        self.g_music_displayer = mpd.MusicDisplayer()
        self.practise_box.pack_start(self.g_music_displayer, True, True, 0)

        self.g_new = gu.bButton(self.action_area, _("_New"), self.new_question)
        gu.bButton(self.action_area, _("440h_z"), self.play_440hz)
        self.g_play_answer = gu.bButton(self.action_area, _("_Play answer"),
            lambda w: self.run_exception_handled(self.m_t.m_P.play_question_arpeggio))
        self.practise_box.show_all()
        ###############
        # config_grid #
        ###############
        self.add_random_transpose_gui(row=0)

    def new_question(self, widget=None):
        def exception_cleanup():
            soundcard.synth.stop()
            self.g_play_answer.set_sensitive(False)
            self.g_music_displayer.clear(2)
        try:
            self.m_t.new_question()
            fontsize = self.get_int('config/feta_font_size=20')
            self.g_music_displayer.display(self.m_t.m_P.get_music(), fontsize)
            self.g_play_answer.set_sensitive(True)
            self.m_t.play_440hz()
        except Exception as e:
            if isinstance(e, mpd.MpdException):
                e.m_mpd_varname = 'music'
                e.m_mpd_badcode = self.m_t.m_P.get_question().music.get_err_context(e, self.m_t.m_P)
            if not self.standard_exception_handler(e, exception_cleanup):
                raise

    def on_start_practise(self):
        super(Gui, self).on_start_practise()
        self.g_music_displayer.clear(2)
        self.g_random_transpose.set_text(str(self.m_t.m_P.header.random_transpose))
        self.g_new.set_sensitive(True)
        self.g_new.grab_focus()
        self.g_play_answer.set_sensitive(False)

    def on_end_practise(self):
        self.m_t.end_practise()
        self.g_play_answer.set_sensitive(False)
        self.g_music_displayer.clear(2)

    def play_440hz(self, widget):
        try:
            self.m_t.play_440hz()
        except Exception as e:
            if not self.standard_exception_handler(e):
                raise



# GNU Solfege - free ear training software
# Copyright (C) 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import re
import sys

from solfege import buildinfo
from solfege import filesystem


def win32_put_langenviron(lang):
    """
    Write the filesystem.app_data()/langenviron.txt file.
    win32 only.
    """
    langfile = open(os.path.join(filesystem.app_data(), "langenviron.txt"), 'w')
    print("# rem Created by GNU Solfege %s" % buildinfo.VERSION_STRING, file=langfile)
    if lang:
        print(lang, file=langfile)
    else:
        print("# System default language. Not setting variable.", file=langfile)
    langfile.close()


def win32_get_langenviron():
    """
    Return the language defined in filesystem.app_data()/langenviron.bat
    win32 only
    """
    assert sys.platform == 'win32'
    fn = os.path.join(filesystem.app_data(), "langenviron.txt")
    lang = None
    if os.path.isfile(fn):
        try:
            for line in open(fn, 'r').readlines():
                if not line.startswith("#"):
                    lang = line.strip()
        except IOError as e:
            # Try here too, just to be sure no unicode shit bothers us.

            try:
                print("IOError reading %s:" % fn, e)
            except Exception:
                pass
    else:
        lang = _pre_3_11_win32_get_langenviron()
    if lang:
        return lang
    else:
        return "system default"


def _pre_3_11_win32_get_langenviron():
    """
    Old version of win32_get_langenviron.
    Return the language defined in filesystem.app_data()/langenviron.bat
    Return None if no language is defined or the file does not exist.
    win32 only
    """
    assert sys.platform == 'win32'
    s = ""
    try:
        langfile = open(os.path.join(filesystem.app_data(), "langenviron.bat"), 'r')
        s = langfile.read()
        langfile.close()
    except IOError:
        # we get here for example when the file does not exist
        return None
    if s:
        r = re.compile("set LANGUAGE=(?P<lang>.*)")
        for line in s.split("\n"):
            m = r.match(line)
            if m:
                return m.groups()[0]
    return None



# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2006, 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import Gtk
from gi.repository import Gdk

import math

from solfege import cfg
from solfege import gu
from solfege import mpd
from solfege import utils


class CairoCommon(object):
    mark_color = {
        1: (0, 1, 0),
        2: (0.8, 0.6, 0)
    }

    def __init__(self):
        self.m_marks = []
        self.m_clicked_tones = []

    def clear(self):
        self.m_marks = []
        self.m_clicked_tones = []

    def mark_note(self, midi_int, color):
        """
        color should be 1 to mark the first tone of a question and
        2 to mark the second tone.
        """
        if midi_int not in self.m_marks:
            self.m_marks.append((midi_int, color))
        self.queue_draw()

    def set_first_note(self, note):
        self.clear()
        self.m_clicked_tones = [int(note)]
        self.mark_note(self.m_clicked_tones[-1], 1)

    def grab_focus_first_sensitive_button(self):
        """
        Dummy function. Only buttons interface implement this.
        """
        pass

    def know_directions(self):
        """
        Return True because this is a inputwidget where the user say
        both interval type _and_ direction.
        """
        return True

    def forget_last_tone(self):
        self.m_clicked_tones.pop()

    def _handle_tone_clicked(self, midi_int, mouse_button):
        if self.m_clicked_tones:
            interval = midi_int - self.m_clicked_tones[-1]
            if mouse_button == 1:
                self.m_clicked_tones.append(midi_int)
            self.m_callback(mouse_button, interval, midi_int)


class PianoKeyboard(Gtk.DrawingArea, CairoCommon):

    def __init__(self, num_octaves, lowest_c, key_w=15):
        Gtk.DrawingArea.__init__(self)
        CairoCommon.__init__(self)
        self.connect("draw", self.draw)
        self.connect("button-press-event", self._on_button_press)
        self.set_events(Gdk.EventMask.BUTTON_PRESS_MASK)
        # Piano stuff
        self.m_num_octaves = num_octaves
        self.m_lowest_c = mpd.MusicalPitch.new_from_notename(lowest_c)
        self.m_lowest_tone = self.m_lowest_c.get_octave_notename()
        self.m_highest_tone = (self.m_lowest_c.clone() + 12 * num_octaves - 1).get_octave_notename()
        self.m_white_h = key_w * 3.4
        self.m_black_h = key_w * 2.0
        self.m_key_w = key_w
        self.m_last_redraw = 0
        self.m_black_w = 0.6
        self.set_size_request(int(num_octaves * 7 * key_w + 1), int(self.m_white_h + 1))

    def _on_button_press(self, drawingarea, event):
        assert event.x >= 0
        if event.x < self.m_pos_x or event.x > self.m_num_octaves * 7 * self.m_key_w + self.m_pos_x:
            # Click outside the keys are ignored.
            return
        click_x = event.x - self.m_pos_x
        octave_w = 7 * self.m_key_w
        clicked_octave = int(click_x / octave_w)
        clicked_white = int((click_x % octave_w) / self.m_key_w)
        black_clicked = 0
        if event.y < self.m_black_h:
            if clicked_white in (0, 1, 3, 4, 5):
                # How many pixels into the white key does the black
                # key start:
                bpos1 = self.m_key_w - self.m_key_w * self.m_black_w / 2
                if click_x % self.m_key_w >= bpos1:
                    black_clicked = 1
            if clicked_white in (1, 2, 4, 5, 6):
                # How many pixels into the white key does the black
                # key end:
                bpos2 = self.m_key_w * self.m_black_w / 2
                if click_x % self.m_key_w <= bpos2:
                    black_clicked = -1
        clicked_on = self.m_lowest_c.clone()
        clicked_on.m_notename_i = clicked_white
        clicked_on.m_accidental_i = black_clicked
        clicked_on.m_octave_i += clicked_octave
        self.on_button_press_event(event, clicked_on.semitone_pitch())

    def draw(self, widget, context):
        num_key = 7 * self.m_num_octaves
        width = num_key * self.m_key_w
        self.m_pos_x = pos_x = int((self.get_allocated_width() - width) / 2) + 0.5
        pos_y = .5
        context.set_line_width(1.2)
        context.save()
        for n in range(num_key):
            context.rectangle(pos_x + n * self.m_key_w, pos_y,
                              self.m_key_w, self.m_white_h)
        context.set_source_rgb(1, 1, 1)
        context.fill_preserve()
        context.set_source_rgb(0, 0, 0)
        context.stroke()
        context.restore()

        # Draw black keys
        context.save()
        for oct in range(self.m_num_octaves):
            for n in (0, 1, 3, 4, 5):
                context.rectangle(pos_x + oct * 7 * self.m_key_w + (n + 1 - self.m_black_w / 2) * self.m_key_w,
                            pos_y,
                            self.m_key_w * self.m_black_w, self.m_black_h)
                context.set_source_rgb(0, 0, 0)
                context.fill_preserve()
        context.stroke()
        context.restore()
        for midi_int, color in self.m_marks:
            n = mpd.MusicalPitch.new_from_int(midi_int)
            acc = n.m_accidental_i
            step = n.steps() - self.m_lowest_c.steps()
            if acc == 0:
                context.arc(pos_x + (step + 0.5) * self.m_key_w, self.m_white_h - (self.m_white_h - self.m_black_h) / 2,
                            self.m_key_w * 0.35, 0, 2 * math.pi)
            else:
                context.arc(pos_x + (step + 0.5 + 0.5 * acc) * self.m_key_w, self.m_black_h * 0.5,
                            self.m_key_w * 0.25, 0, 2 * math.pi)
            context.set_source_rgb(*self.mark_color[color])
            context.fill_preserve()
            context.stroke()


class PianoOctaveWithAccelName(PianoKeyboard):

    def __init__(self, callback, keys):
        PianoKeyboard.__init__(self, 1, "c", 40)
        self.m_callback = callback
        self.m_visible_accels = False
        self.m_keys = keys

    def draw(self, widget, context):
        PianoKeyboard.draw(self, widget, context)
        if not self.m_visible_accels:
            return
        context.save()
        context.select_font_face("Sans")
        context.set_font_size(24)
        text_h = context.text_extents('A')[3]
        for idx, n in enumerate((0, 2, 4, 5, 7, 9, 11)):
            context.new_path()
            text_w = context.text_extents(self.m_keys[n])[4]
            context.move_to(self.m_pos_x + idx * self.m_key_w + self.m_key_w / 2 - text_w / 2, self.m_black_h + (self.m_white_h - self.m_black_h) / 2 + text_h / 2)
            context.text_path(self.m_keys[n])
            context.fill()
            context.stroke()
        context.restore()
        context.save()
        context.select_font_face("Sans")
        context.set_font_size(24)
        for idx, n in enumerate((1, 3, None, 6, 8, 10)):
            if n:
                context.new_path()
                text_w = context.text_extents(self.m_keys[n])[4]
                context.move_to(self.m_pos_x + (idx + 1) * self.m_key_w - text_w / 2, self.m_black_h * 0.8)
                context.text_path(self.m_keys[n])
                context.set_source_rgb(1, 1, 1)
                context.fill()
                context.stroke()
        context.restore()

    def on_button_press_event(self, event, midi_int):
        if event.button == 3:
            utils.play_note(4, midi_int)
        elif event.button == 1:
            self.m_callback(mpd.int_to_octave_notename(midi_int))


class IntervalPianoWidget(PianoKeyboard):

    def __init__(self, callback):
        PianoKeyboard.__init__(self, 4, "c,", 18)
        self.m_callback = callback

    def on_button_press_event(self, event, midi_int):
        """
        The callback function is only called if we have an interval.
        """
        self._handle_tone_clicked(midi_int, event.button)


class IntervalButtonsWidgetBase(Gtk.Grid, cfg.ConfigUtils):
    use_users_vocal_range = True

    def __init__(self, exname, callback, sensicallback):
        """
        callback -- the function to call when the user click on a button.
        sensicallback -- a function that will return a list of all interval
                         it will be asked for.
        """
        Gtk.Grid.__init__(self)
        cfg.ConfigUtils.__init__(self, exname)
        self.set_orientation(Gtk.Orientation.VERTICAL)
        self.get_sensitive_buttons = sensicallback
        self.m_callback = callback
        self.m_buttons = {}

    def post_construct(self, vars_to_watch):
        self.m_lowest_tone = mpd.LOWEST_NOTENAME
        self.m_highest_tone = mpd.HIGHEST_NOTENAME
        self.add_watch('disable_unused_intervals', self.intervals_changed)
        for var in vars_to_watch:
            self.add_watch(var, self.intervals_changed)
        self.intervals_changed()

    def intervals_changed(self, s=None):
        if self.get_bool('disable_unused_intervals'):
            self.set_sensitivity(self.get_sensitive_buttons())
        else:
            self.set_sensitivity(list(range(mpd.interval.max_interval + 1)))

    def set_sensitivity(self, make_active):
        for x in range(1, mpd.interval.max_interval + 1):
            if x in self.m_buttons:
                self.m_buttons[x].set_sensitive(x in make_active)

    def mk_button(self, nr):
        # buttonwidget calls m_callback with None as midi_int because it
        # does not know if you mean interval up or down when you click
        # the buttons
        txt = mpd.Interval.new_from_int(nr).get_name()
        self.m_buttons[nr] = b = Gtk.Button(txt)
        b.connect('clicked',
                  lambda s, nr=nr, self=self: self.m_callback(1, nr, None))
        b.connect('button-press-event', self._abc, nr)
        b.set_vexpand(True)
        b.set_hexpand(True)
        return b

    def row(self, *data):
        row = Gtk.Grid()
        row.set_column_homogeneous(True)
        self.add(row)
        for i in data:
            row.add(self.mk_button(i))

    def _abc(self, button, event, interval):
        if event.button == 3:
            self.m_callback(3, interval, None)

    def set_first_note(self, note):
        self.m_first_note = int(note)

    def know_directions(self):
        return False

    def clear(self):
        pass

    def show(self):
        self.show_all()

    def grab_focus_first_sensitive_button(self):
        if self.get_bool('disable_unused_intervals'):
            self.m_buttons[self.get_sensitive_buttons()[0]].grab_focus()
        else:
            self.m_buttons[1].grab_focus()

    def mark_note(self, midi_int, color):
        """
        Only implemented for buttons that are drawn using cairo.
        """
        pass


class IntervalButtonsWidget(IntervalButtonsWidgetBase):

    def __init__(self, exname, callback, sensicallback, vars_to_watch):
        IntervalButtonsWidgetBase.__init__(self, exname, callback, sensicallback)
        g1 = Gtk.Grid()
        for i, (x, y) in enumerate((
                (0, 0), (1, 0), (2, 0), (3, 0),
                (0, 1), (1, 1), (2, 1), (3, 1),
                (0, 2), (1, 2), (2, 2), (3, 2))):
            g1.attach(self.mk_button(i + 1), x, y, 1, 1)
        g2 = Gtk.Grid()
        for i, (x, y) in enumerate((
                (0, 0), (1, 0), (2, 0), (3, 0),
                (0, 1), (1, 1), (2, 1), (3, 1),
                (0, 2), (1, 2), (2, 2), (3, 2))):
            g2.attach(self.mk_button(i + 13), x, y, 1, 1)
        self.set_row_spacing(gu.hig.SPACE_MEDIUM)
        self.add(g1)
        self.add(g2)
        self.post_construct(vars_to_watch)


class IntervalButtonsWidget2(IntervalButtonsWidgetBase):

    def __init__(self, exname, callback, sensicallback, vars_to_watch):
        IntervalButtonsWidgetBase.__init__(self, exname, callback, sensicallback)
        self.row(1, 2, 3, 4)
        self.row(5, 6, 7)
        self.row(8, 9, 10, 11)
        self.row(12)
        self.row(13, 14, 15, 16)
        self.row(17, 18, 19)
        self.row(20, 21, 22, 23)
        self.row(24)
        self.post_construct(vars_to_watch)


class AbstractGuitarWidget(Gtk.DrawingArea, CairoCommon):

    def __init__(self, callback, strings,
                string_thickness=(1, 1, 1, 1, 1, 1)):
        Gtk.DrawingArea.__init__(self)
        CairoCommon.__init__(self)
        self.m_callback = callback
        self.m_strings = strings
        self.m_string_thickness = string_thickness
        self.m_fretdist = (20, 39, 38, 37, 36, 35, 34, 33, 32, 31, 30, 30, 30)
        self.m_stringdist = 17
        self.m_numstring = len(self.m_strings)
        self.m_neckborder = 6
        self.m_neckl = 0
        self.m_xlist = []
        self.m_lowest_tone = mpd.int_to_octave_notename(
            min(list(map(mpd.notename_to_int, self.m_strings))))
        self.m_highest_tone = mpd.int_to_octave_notename(
            len(self.m_fretdist) - 1 + max(list(map(mpd.notename_to_int, self.m_strings))))
        self.m_stringtuning = list(map(mpd.notename_to_int, self.m_strings))
        tmp = 0
        for x in self.m_fretdist:
            tmp = tmp + x
            self.m_xlist.append(tmp)
        tmp = self.m_neckborder + self.m_stringdist / 2
        self.m_ylist = [tmp]
        for y in range(self.m_numstring - 1):
            tmp = tmp + self.m_stringdist
            self.m_ylist.append(tmp)
        for x in self.m_fretdist:
            self.m_neckl = self.m_neckl + x
        self.m_neckl = self.m_neckl + 2
        self.m_neckw = self.m_neckborder \
                       + (self.m_numstring - 1) * self.m_stringdist \
                       + 1 + self.m_neckborder
        self.set_events(Gdk.EventMask.BUTTON_PRESS_MASK | Gdk.EventMask.BUTTON_RELEASE_MASK | Gdk.EventMask.POINTER_MOTION_MASK | Gdk.EventMask.ENTER_NOTIFY_MASK | Gdk.EventMask.LEAVE_NOTIFY_MASK)
        self.connect('draw', self.draw)
        self.m_mouse_pos = None, None
        self.connect('motion-notify-event', self.on_motion_notify_event)
        self.connect('leave-notify-event', self.on_leave_notify_event)
        self.connect('button-press-event', self.on_button_press_event)
        self.set_size_request(self.m_neckl, self.m_neckw)

    def on_button_press_event(self, widget, event):
        x, y = self.event2xy(event)
        if x is not None and y is not None:
            midi_int = self.m_stringtuning[y] + x
            self._handle_tone_clicked(midi_int, event.button)

    def on_leave_notify_event(self, widget, event):
        self.m_mouse_pos = None, None
        self.queue_draw()

    def on_motion_notify_event(self, widget, event):
        x, y = self.event2xy(event)
        need_redraw = False
        if x is not None and y is not None:
            if (x, y) != self.m_mouse_pos:
                need_redraw = True
            self.m_mouse_pos = x, y
        else:
            if (x, y) != self.m_mouse_pos:
                need_redraw = True
            self.m_mouse_pos = None, None
        if need_redraw:
            self.queue_draw()

    def draw(self, widget, context):
        self.m_posx = int((self.get_allocated_width() - self.m_neckl) / 2) + 0.5
        self.m_posy = int((self.get_allocated_height() - self.m_neckw) / 2) + 0.5
        context.set_line_width(1.2)
        context.save()
        LG = 0.75  # white grey
        DG = 0.66  # dark grey
        # fret board
        context.rectangle(self.m_posx, self.m_posy, self.m_neckl, self.m_neckw)
        context.fill_preserve()
        context.stroke()
        context.restore()
        # first fret
        context.save()
        px = self.m_fretdist[0] + self.m_posx
        context.rectangle(px, self.m_posy,
                          5, self.m_neckw)
        context.set_source_rgb(LG, LG, LG)
        context.fill_preserve()
        context.set_source_rgb(LG, LG, LG)
        context.stroke()
        #
        context.set_source_rgb(1, 1, 1)
        context.move_to(px, self.m_posy)
        context.rel_line_to(0, self.m_neckw)
        context.stroke()
        #
        context.set_source_rgb(DG, DG, DG)
        context.move_to(px + 5, self.m_posy)
        context.rel_line_to(0, self.m_neckw)
        context.stroke()
        context.restore()
        #
        # FRETS
        context.save()
        for w in self.m_fretdist[1:]:
            px = px + w
            context.set_source_rgb(1, 1, 1)
            context.move_to(px, self.m_posy)
            context.rel_line_to(0, self.m_neckw)
            context.stroke()
            context.set_source_rgb(LG, LG, LG)
            context.move_to(px + 1, self.m_posy)
            context.rel_line_to(0, self.m_neckw)
            context.stroke()
            context.set_source_rgb(DG, DG, DG)
            context.move_to(px + 2, self.m_posy)
            context.rel_line_to(0, self.m_neckw)
            context.stroke()
        context.restore()
        #
        # String
        context.save()
        for y in range(self.m_numstring):
            if self.m_string_thickness[y] == 1:
                context.set_line_width(self.m_string_thickness[y])
                context.set_source_rgb(LG, LG, LG)
                context.move_to(self.m_posx, self.m_posy + self.m_neckborder
                                + y * self.m_stringdist)
                context.rel_line_to(self.m_neckl, 0)
                context.stroke()
            if self.m_string_thickness[y] == 2:
                context.set_source_rgb(1, 1, 1)
                context.move_to(self.m_posx, self.m_posy + self.m_neckborder
                                + y * self.m_stringdist)
                context.rel_line_to(self.m_neckl, 0)
                context.stroke()
                context.set_source_rgb(LG, LG, LG)
                context.move_to(self.m_posx, self.m_posy + self.m_neckborder
                                + y * self.m_stringdist + 1)
                context.rel_line_to(self.m_neckl, 0)
                context.stroke()
            if self.m_string_thickness[y] == 3:
                context.rectangle(self.m_posx, self.m_posy + self.m_neckborder
                                + y * self.m_stringdist + 1,
                                self.m_neckl, 1)
                context.set_source_rgb(LG, LG, LG)
                context.fill_preserve()
                context.set_source_rgb(LG, LG, LG)
                context.stroke()
                context.set_source_rgb(1, 1, 1)
                context.move_to(self.m_posx, self.m_posy + self.m_neckborder
                                + y * self.m_stringdist)
                context.rel_line_to(self.m_neckl, 0)
                context.stroke()
        context.restore()
        if isinstance(self, GuitarWidget):
            # The white dots on a guitar
            context.save()
            for x, y in ((3, 2), (5, 2), (7, 2), (9, 2), (12, 1), (12, 3)):
                context.arc(self.m_posx + self.m_xlist[x] - self.m_fretdist[x] / 2,
                            self.m_posy + self.m_ylist[y],
                            4.5, 0, math.pi * 2)
                context.set_source_rgb(1, 1, 1)
                context.fill_preserve()
                context.stroke()
            context.restore()
        # Marks
        context.save()
        for note, color in self.m_marks:
            for idx in range(len(self.m_stringtuning)):
                if self.m_stringtuning[idx] <= note < self.m_stringtuning[idx] + len(self.m_fretdist):
                    x = note - self.m_stringtuning[idx]
                    context.arc(self.m_posx + self.m_xlist[x] - self.m_fretdist[x] / 2,
                        self.m_posy + self.m_ylist[idx] - self.m_stringdist / 2,
                        7, 0, math.pi * 2)
                    context.set_source_rgb(*self.mark_color[color])
                    context.fill_preserve()
                    context.set_source_rgb(0, 0, 0)
                    context.stroke()
        context.restore()
        # Dot following mouse cursor
        if self.m_mouse_pos != (None, None):
            context.save()
            context.arc(self.m_posx + self.m_xlist[self.m_mouse_pos[0]] - self.m_fretdist[self.m_mouse_pos[0]] / 2,
                        self.m_posy + self.m_ylist[self.m_mouse_pos[1]] - self.m_stringdist / 2,
                        5, 0, math.pi * 2)
            context.set_source_rgb(1, 0, 0)
            context.fill_preserve()
            context.set_source_rgb(0, 0, 0)
            context.stroke()
            context.restore()

    def event2xy(self, event):
        x = event.x - self.m_posx
        xp = yp = None
        for idx in range(len(self.m_xlist)):
            if 0 <= x < self.m_xlist[idx]:
                xp = idx
                break
        for idx in range(len(self.m_ylist)):
            if 0 <= event.y < self.m_ylist[idx]:
                yp = idx
                break
        return xp, yp


class GuitarWidget(AbstractGuitarWidget):

    def __init__(self, callback, strings, string_thickness):
        AbstractGuitarWidget.__init__(self, callback, strings, string_thickness)


class AccordionWidget(Gtk.DrawingArea, CairoCommon):

    def __init__(self, callback, keyboard_system):
        Gtk.DrawingArea.__init__(self)
        CairoCommon.__init__(self)
        self.m_callback = callback

        def ff(first, count):
            i = mpd.notename_to_int(first)
            return [mpd.int_to_octave_notename(i + x * 3) for x in range(count)]
        keyboard_def = {'norwegian':
            (ff("g,", 19), ff("f,", 20), ff("fis,", 19),
            ff("e,", 20), ff("f,", 19))}
        keyboard_def['swedish'] = (
            ff("e,", 19), ff("dis,", 20), ff("f,", 19),
            ff("e,", 20), ff("fis,", 19))
        keyboard_def['finnish'] = (
            ff("dis,", 19), ff("d,", 20), ff("e,", 19),
            ff("dis,", 20), ff("f,", 19))
        keyboard_def['belgian'] = (
            ff("fis,", 19), ff("e,", 20), ff("f,", 19),
            ff("es,", 20), ff("e,", 19)
        )
        self.m_notenames = keyboard_def[keyboard_system]
        self.m_lowest_tone = mpd.HIGHEST_NOTENAME
        self.m_highest_tone = mpd.LOWEST_NOTENAME
        for v in self.m_notenames:
            if mpd.compare_notenames(self.m_lowest_tone, v[0]) > 0:
                self.m_lowest_tone = v[0]
            if mpd.compare_notenames(self.m_highest_tone, v[-1]) < 0:
                self.m_highest_tone = v[-1]
        self.m_button_radius = 9
        self.m_button_xdist = 20
        self.m_button_ydist = 18

        self.connect('draw', self.draw)
        self.connect('button-press-event', self._on_button_press)
        self.set_events(Gdk.EventMask.BUTTON_PRESS_MASK)

    def event_to_button_pos(self, event):
        """
        Return a tuple telling which button on the accordion that
        got clicked.
        Returns (Col. row)
        """
        for y in range(self.m_by):
          if self.m_bb + self.m_posy + y * self.m_button_ydist - self.m_button_radius < event.y < self.m_bb + self.m_posy + (y + 1) * self.m_button_ydist - self.m_button_radius:
            for x in range(self.m_bx):
                # button center
                cx = self.m_bb + self.m_posx + x * self.m_button_xdist + (y + 1) % 2 * self.m_button_radius
                cy = self.m_bb + self.m_posy + y * self.m_button_ydist
                if math.sqrt((event.x - cx)**2 + (event.y - cy)**2) < self.m_button_radius:
                    return x, y
        return None, None

    def _on_button_press(self, drawingarea, event):
        # This is only set if set_first_note is called
        x, y = self.event_to_button_pos(event)
        if x is None:
            return
        midi_int = mpd.notename_to_int(self.m_notenames[y][x])
        self._handle_tone_clicked(midi_int, event.button)

    def draw(self, widget, context):
        # number of buttons, horiz and vertic.
        self.m_bx = bx = len(self.m_notenames[0])
        self.m_by = by = len(self.m_notenames)
        # how much space between buttons and border
        self.m_bb = bb = 16
        #
        w = bx * self.m_button_xdist + bb * 2
        h = by * self.m_button_ydist + bb * 2
        self.m_posx = posx = int((self.get_allocated_width() - w) / 2) + 0.5
        self.m_posy = posy = int((self.get_allocated_height() - h) / 2) + 0.5
        context.set_line_width(1.2)
        context.save()
        context.move_to(posx, posy)
        for x, y in ((w, 0),
                     (0, self.m_button_ydist * by),  # right btn
                     (-self.m_button_xdist, self.m_button_ydist),  # right btn
                     (-w + self.m_button_xdist * 2, 0),
                     (-self.m_button_xdist, -self.m_button_ydist),
                     (0, -self.m_button_ydist * by),
                     ):
            context.rel_line_to(x, y)
        context.close_path()
        context.set_source_rgb(0, 0, 0)
        context.fill_preserve()
        context.stroke()
        context.restore()

        # buttons
        context.save()
        for row_idx, row in enumerate(self.m_notenames):
            for col_idx, notename in enumerate(row):
                context.arc(bb + posx + col_idx * self.m_button_xdist + (row_idx + 1) % 2 * self.m_button_radius,
                            bb + posy + row_idx * self.m_button_ydist,
                            self.m_button_radius, 0, 2 * math.pi)
                if notename[1:][:2] in ('is', 'es'):
                    context.set_source_rgb(0.4, 0.4, 0.4)
                else:
                    context.set_source_rgb(1, 1, 1)
                context.fill_preserve()
                context.stroke()
        context.restore()

        context.save()
        for row_idx, row in enumerate(self.m_notenames):
            for col_idx, notename in enumerate(row):
                try:
                    mark_idx = [m[0] for m in self.m_marks].index(mpd.notename_to_int(notename))
                except ValueError:
                    mark_idx = -1

                if mark_idx != -1:
                    context.arc(bb + posx + col_idx * self.m_button_xdist + (row_idx + 1) % 2 * self.m_button_radius,
                            bb + posy + row_idx * self.m_button_ydist,
                            5, 0, 2 * math.pi)
                    context.set_source_rgb(*self.mark_color[self.m_marks[mark_idx][1]])
                    context.fill_preserve()
                    context.stroke()

        context.restore()
        self.set_size_request(500, self.m_by * self.m_button_ydist + 2 * self.m_bb)


inputwidget_names = [_("Buttons"),
                 _("Buttons %i") % 2,
                 _("Piano"), _("Guitar"), _("Bass"),
                 _("5 string bass"), _("6 string bass"),
                 _("Accordion B griff"), _("Accordion C griff"),
                 _("Accordion (system used in Finland)"),
                 _("Accordion (Belgian Charleroi do2)")]


def int_to_inputwidget(i, callback, exname, get_interval_input_list, vars_to_watch):
    if i in (0, 1):
        w = {0: IntervalButtonsWidget,
             1: IntervalButtonsWidget2}[i](exname, callback,
                                           get_interval_input_list,
                                           vars_to_watch)
    elif i == 2:
        w = IntervalPianoWidget(callback)
    elif i == 3:
        w = GuitarWidget(callback, ["e'", "b", "g", "d", "a,", "e,"],
                                   (1, 1, 1, 2, 2, 3))
    elif i == 4:
        w = AbstractGuitarWidget(callback, ["g,", "d,", "a,,", "e,,"],
                                 (1, 1, 2, 3))
    elif i == 5:
        w = AbstractGuitarWidget(callback, ["g,", "d,", "a,,", "e,,", "b,,,"],
                                 (1, 1, 2, 3, 3))
    elif i == 6:
        w = AbstractGuitarWidget(callback, ["c", "g,", "d,", "a,,", "e,,", "b,,,"],
                                 (1, 1, 2, 2, 3, 3))
    elif i == 7:
        w = AccordionWidget(callback, 'norwegian')
    elif i == 8:
        w = AccordionWidget(callback, 'swedish')
    elif i == 9:
        w = AccordionWidget(callback, 'finnish')
    elif i == 10:
        w = AccordionWidget(callback, 'belgian')
    return w



# GNU Solfege - free ear training software
# Copyright (C) 2005, 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
This file does the checking for optional features that depend on
python modules the user can have installed.

It also does sanity check on the python and gtk versions and some
other initial setup tasks.
"""

import sys
import os


def assert_python_version(required_version):
    if sys.version_info < required_version:
        sys.exit("Solfege need Python %s or newer. The configure script told you so.\nThis is Python %s" % (".".join([str(i) for i in required_version]), sys.version))


def init(options):
    # this is needed for py2exe
    if sys.platform == 'win32':
        os.environ['PATH'] += ";lib;bin;"
    from gi.repository import Gtk
    assert_python_version((2, 7))
    Gtk.check_version(3, 4, 0)
